import tempfile

import os
import shutil
from tko.util.text import Text
from tko.util.runner import Runner
from tko.settings.settings import Settings
import yaml #type: ignore
from pathlib import Path
from tko.util.decoder import Decoder
import io

class CompileError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message

class Executable:
    def __init__(self, cmd: list[str] | str | None, files: list[Path] | None, folder: Path):
        if cmd is None:
            cmd = []
        if files is None:
            files = []
        self.__cmd_list: list[str] | str = cmd
        self.__folder: Path = folder
        self.__compiled: bool = False
        self.__compile_error: bool = False
        self.__error_msg: Text = Text()
        self.__shell_mode: bool = False # subprocess needs bash mode to process symbols like & or |
    
    def needs_shell_mode(self):
        return self.__shell_mode
    
    def set_executable(self, cmd: list[str] | str, files: list[Path], folder: Path, shell_mode: bool):
        self.__compiled = True
        self.__cmd_list = cmd
        self.__files: list[Path] = files
        self.__folder: Path = folder.resolve()
        self.__shell_mode = shell_mode
        return self
    
    def get_command(self) -> tuple[list[str] | str, Path]:
        cmd: list[str] | str = self.__cmd_list
        if isinstance(cmd, str):
            cmd += " " + " ".join([str(file.resolve()) for file in self.__files])
        else:
            cmd += [str(file.resolve()) for file in self.__files]
        return cmd, self.__folder

    def set_compile_error(self, error_msg: Text | str):
        self.__compiled = True
        self.__compile_error = True
        if isinstance(error_msg, str):
            self.__error_msg = Text().add(error_msg)
        else:
            self.__error_msg = error_msg
        return self
    
    def compiled(self):
        return self.__compiled

    def has_compile_error(self):
        return self.__compiled and self.__compile_error
    
    def get_error_msg(self):
        return self.__error_msg

class SolverBuilder:
    def __init__(self, args_list: list[Path], settings: Settings):
        self.settings = settings
        self.args_list: list[Path] = args_list
        self.cache_dir: Path = Path()
        if len(self.args_list) > 0:
            self.cache_dir = self.args_list[0].parent / ".build"
        else:
            self.cache_dir = Path(tempfile.mkdtemp())
        self.clear_cache()
        self.__exec: Executable = Executable(cmd=None, files=None, folder=Path(""))

    def check_tool(self, name: str):
        if shutil.which(name) is None:
            self.__exec.set_compile_error(Text.format("{r}: O comando '" + name + "' não foi encontrado", "Falha")) 
            raise CompileError("fail: comando '" + name + "' não foi encontrado")

    def not_compiled(self):
        return not self.__exec.compiled()

    def has_compile_error(self):
        return self.__exec.has_compile_error()

    def set_main(self, main: str):
        list_main: list[Path] = []
        list_other: list[Path] = []

        for path in self.args_list:
            if path.name == main:
                list_main.append(path)
            else:
                list_other.append(path)

        self.args_list = list_main + list_other
        return self

    def clear_cache(self):
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir, ignore_errors=True)
        os.makedirs(self.cache_dir, exist_ok=True)

    def get_executable(self, force_rebuild: bool = False) -> tuple[Executable, bool]:
        if not self.args_list:
            return Executable(cmd=None, files=None, folder=Path("")), False
        if self.__exec.has_compile_error():
            return self.__exec, False
        if not self.__exec.compiled() or force_rebuild:
            self.prepare_exec()
        return self.__exec, True

    def reset(self):
        self.__exec = Executable(cmd=None, files=None, folder=Path(""))
        self.clear_cache()

    def prepare_exec(self) -> None:
        self.reset()
        first = self.args_list[0]

        if first.suffix == ".mk":
            self.__prepare_make()
        elif first.suffix == ".yaml":
            self.__prepare_yaml()
        elif first.suffix == ".ts":
            self.__prepare_ts()
        elif first.suffix[1:] in self.settings.languages.get_languages().keys():
            self.prepare_exec_with_lang()
        else:
            self.__exec.set_executable([str(x) for x in self.args_list], [], Path(""), shell_mode=True)

    def __prepare_yaml(self):
        solver: Path = self.args_list[0]
        folder = solver.parent
        content = Decoder.load(solver)
        yaml_data = yaml.safe_load(content)

        if "build" in yaml_data and yaml_data["build"] is not None:
            build_txt = yaml_data["build"]
            if not isinstance(build_txt, str):
                raise Warning(Text.format("{r}: O campo build deve ser uma string", "Falha"))
            
            return_code, stdout, stderr = Runner.subprocess_run(cmd=build_txt, folder=folder, shell_mode=True)
            if return_code != 0:
                self.__exec.set_compile_error(stdout + stderr)
                return
        if not "run" in yaml_data:
            raise Warning(Text.format("{r}: Seu arquivo yaml precisa ter um campo {g} ", "Falha", "run:"))
        if not isinstance(yaml_data["run"], str):
            raise Warning(Text.format("{r}: O campo run deve ser uma string", "Falha"))
        self.__exec.set_executable(cmd=yaml_data["run"], files=[], folder=folder, shell_mode=True)

    def replace_placeholders(self, text: str) -> str:
        parent_folder = self.args_list[0].parent
        main_file_without_ext = self.args_list[0].stem
        exe_ext = ".exe" if os.name == "nt" else ""
        output_path = self.cache_dir / ("a.out" + exe_ext)
        
        files_str = " ".join([f'"{str(x.relative_to(parent_folder))}"' for x in self.args_list])
        text = (text.replace("{files}", files_str)
                    .replace("{output}", f'"{str(output_path.relative_to(parent_folder))}"')
                    .replace("{main}", main_file_without_ext)
                    .replace("{cache}", f'"{str(self.cache_dir.relative_to(parent_folder))}"')).strip()

        return text

    def prepare_exec_with_lang(self):
        lang = self.settings.languages.get_languages().get(self.args_list[0].suffix[1:], None)
        if lang is None:
            self.__exec.set_compile_error(Text.format("{r}: Extensão de arquivo '" + self.args_list[0].suffix + "' não reconhecida e sem configuração de linguagem", "Falha"))
            return
        build_cmd = lang.build_cmd
        run_cmd = lang.run_cmd
        parent_folder = self.args_list[0].parent
        build_cmd = self.replace_placeholders(build_cmd)
        return_code, stdout, stderr = Runner.subprocess_run(build_cmd, folder=parent_folder, shell_mode=True)
        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
            return
        run_cmd = self.replace_placeholders(run_cmd)
        self.__exec.set_executable(run_cmd, [], parent_folder, shell_mode=True)

    def __prepare_make(self):
        self.check_tool("make")
        solver = os.path.abspath(self.args_list[0])
        folder = os.path.dirname(solver)
        cmd = ["make", "-s", "-C", folder, "-f", solver, "build"]
        return_code, stdout, stderr = Runner.subprocess_run(cmd)
        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            self.__exec.set_executable(cmd=["make", "-s", "-C", folder, "-f", solver, "run"], files=[], folder=Path(""), shell_mode=False)

    def update_input_function(self, path_list: list[Path], copy_dir: Path):
        new_files: list[str] = []
        for origin in self.args_list:
            new_files.append(os.path.join(copy_dir, os.path.basename(origin)))

        for i in range(len(path_list)):
            origin = path_list[i]
            destiny = new_files[i]

            content = Decoder.load(origin)
            lines = content.splitlines(keepends=True)
            io_target = io.StringIO()
            tag = 'constinput=()=>""'

            input_cmd = r'const input: () => string = (() => { let lines: string[] | undefined; let index = 0; let readlineSync: any; return (): string => { const isTTY = process?.stdin?.isTTY; if (isTTY) { readlineSync = readlineSync ?? require("readline-sync"); return readlineSync.question(); } if (!lines) { try { const fs = require("fs"); lines = fs.readFileSync(0, "utf-8").split(/\r?\n/); } catch { lines = []; } } return lines![index++] ?? ""; }; })();'

            inserted: bool = False
            for line in lines:
                match = False
                if not inserted:
                    filtered = "".join([c for c in line if c != " "])
                    match = filtered.startswith(tag)
                if match:
                    inserted = True
                    io_target.write(input_cmd)
                else:
                    io_target.write(line)

            with open(destiny, "w", encoding="utf-8") as f:
                f.write(io_target.getvalue())
            io_target.close()
        return new_files

    def __prepare_ts(self):
        copy_dir = self.cache_dir / "src"
        # remove the cache dir
        if os.path.exists(copy_dir):
            shutil.rmtree(copy_dir, ignore_errors=True)
        os.makedirs(copy_dir, exist_ok=True)
        new_files = self.update_input_function(self.args_list, copy_dir)

        transpiler = "npx"
        if os.name == "nt":
            transpiler += ".cmd"

        self.check_tool(transpiler)
        self.check_tool("node")
        transpiler = ["npx", "esbuild"]
        cmd: list[str] = transpiler + new_files + ["--outdir=" + str(self.cache_dir), "--format=cjs", "--log-level=error"]
        return_code, stdout, stderr = Runner.subprocess_run(cmd)

        if return_code != 0:
            self.__exec.set_compile_error(stdout + stderr)
        else:
            new_source_list: list[Path] = []
            for source in new_files:
                new_source_list.append(self.cache_dir / (os.path.basename(source)[:-2] + "js"))
            self.__exec.set_executable(cmd=["node"], files=new_source_list, folder=Path(""), shell_mode=False)  # renaming solver to main



import tomllib
from pathlib import Path
from tko.util.decoder import Decoder
from platformdirs import user_data_dir

class LangSettings:
    def __init__(self, build_cmd: str, run_cmd: str, draft: str):
        self.build_cmd: str = build_cmd
        self.run_cmd: str = run_cmd
        self.draft: str = draft
    
    def to_dict(self):
        return self.__dict__
    
    def from_dict(self, attr_dict: dict[str, str]):
        self.build_cmd = attr_dict.get("build_cmd", self.build_cmd)
        self.run_cmd = attr_dict.get("run_cmd", self.run_cmd)
        self.draft = attr_dict.get("draft", self.draft)
        return self

draft_zip = r"""
const std = @import("std");

pub fn main() !void {
    try std.fs.File.stdout().writeAll("hello world!\n");
}"""[1:]

draft_rust = r"""
fn main() {
    println!("Hello, World!");
}"""[1:]

draft_go = r"""
package main
import "fmt"
func main() {
    fmt.Println("Hello, World!")
}"""[1:]

draft_js = r"""
const input=(()=>{let l,i=0,P;return()=>process.stdin.isTTY?((P=P||require("readline-sync")).question()):(l=l||require("fs").readFileSync(0,"utf-8").split(/\r?\n/),l[i++])})();
console.log("Hello, World!");
"""

haskell_draft = r"""
main :: IO ()
main = putStrLn "Hello, World!"
"""

c_draft = r"""
#include <stdio.h>
int main() {
    printf("Hello, World!\n");
    return 0;
}
"""

cpp_draft = r"""
#include <iostream>
int main() {
    std::cout << "Hello, World!" << std::endl;
    return 0;
}
"""

java_draft = r"""
public class draft {
    public static void main(String args[]) {
        System.out.println("Hello, World!");
    }
}
"""

class UserLanguages:

    default_lang_settings: dict[str, LangSettings] = {
        "zig": LangSettings(
            build_cmd="zig build-exe {files} -femit-bin={output}",
            run_cmd="{output}",
            draft=draft_zip
        ),
        "rust": LangSettings(
            build_cmd="rustc {files} -o {output}",
            run_cmd="{output}",
            draft=draft_rust
        ),
        "go": LangSettings(
            build_cmd="go build -o {output} {files}",
            run_cmd="{output}",
            draft=draft_go
        ),
        "py": LangSettings(
            build_cmd="",
            run_cmd="python3 {files}",
            draft="# Escreva seu código aqui\nprint('Hello, World!')"
        ),
        "js": LangSettings(
            build_cmd="",
            run_cmd="node {files}",
            draft=draft_js,
        ), 
        "hs": LangSettings(
            build_cmd="ghc {files} -o {output}",
            run_cmd="{output}",
            draft=haskell_draft
        ),
        "c": LangSettings(
            build_cmd="gcc -Wall {files} -o {output} -lm",
            run_cmd="{output}",
            draft=c_draft
        ),
        "cpp": LangSettings(
            build_cmd="g++ -std=c++17 -Wall -Wextra -Werror -Wno-deprecated {files} -o {output}",
            run_cmd="{output}",
            draft=cpp_draft
        ),
        "java": LangSettings(
            build_cmd="javac {files} -d {cache}",
            run_cmd="java -cp {cache} {main}",
            draft=java_draft
        )
    }
    def get_languages(self) -> dict[str, LangSettings]:
        return self.lang_settings

    def __init__(self, path: Path):
        self.path = path
        self.lang_settings: dict[str, LangSettings] = {}

    def set_default_path(self):
        self.path = Path(user_data_dir("tko")) / "languages.toml"
        return self

    def reset(self):
        self.lang_settings = self.default_lang_settings.copy()
        return self

    def save_file_settings(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.path, "w") as f:
            for lang, settings in self.lang_settings.items():
                f.write(f"[{lang}]\n")
                for key, value in settings.to_dict().items():
                    if value == "":
                        f.write(f"{key} = ''\n")
                    else:
                        f.write(f"{key} = '''\n{value.strip()}\n'''\n")
                f.write("\n")
        return self

    def load_file_settings(self):
        if not self.path.exists():
            self.reset()
            self.save_file_settings()
        try:
            content = Decoder.load(self.path)
            data = tomllib.loads(content)
            for lang, settings in data.items():
                self.lang_settings[lang] = LangSettings(
                    build_cmd=settings.get("build_cmd", ""),
                    run_cmd=settings.get("run_cmd", ""),
                    draft=settings.get("draft", "")
                )
        except Exception as e:
            print(f"Erro ao carregar as configurações de linguagem {self.path}, resetando para as configurações padrão. Erro: {e}")
            self.reset()
            self.save_file_settings()
        return self



from .task_info import TaskInfo

class TaskGrader:
    def __init__(self, task_info: TaskInfo):
        self.info = task_info

    def get_percent(self):
        # hardcoded gambi for compatibility
        if not self.info.feedback:
            return 0.0
        rate = float(self.info.rate)
        if self.info.ia_code:
            rate *= 0.3
        if self.info.ia_debug:
            rate *= 0.8
        if self.info.ia_problem:
            rate *= 0.9
        return rate

    def get_ratio(self) -> float:
        return self.get_percent() / 100.0



import json
from diff_match_patch import diff_match_patch # type: ignore
import os

class PatchInfo:
    def __init__(self, label: str, content: str, lines: int | None = None):
        self.label: str = label
        self.content: str = content
        if lines is not None:
            self.lines: int = lines
        else:
            self.lines: int = len(content.splitlines()) if content else 0
    
    def __str__(self):
        return "{}:{}".format(self.label, self.content)


class PatchHistory:
    def __init__(self):
        # the last element is the integral last version
        # all the other elements are patches to be applied
        self.patches: list[PatchInfo] = [] 
        self.json_file: str | None = None

    def set_json_file(self, path: str):
        self.json_file = path
        return self

    def load_json(self):
        if not self.json_file:
            return self
        if os.path.exists(self.json_file):
            try:
                with open(self.json_file, 'r') as file:
                    data = json.load(file)
                    for patch in data["patches"]:
                        if "lines" in patch:
                            lines = int(patch["lines"])
                        else:
                            lines = None
                        self.patches.append(PatchInfo(patch["label"], patch["content"], lines))
            except:
                pass
        return self

    # save patches, label and content to json file
    def save_json(self):
        if not self.json_file:
            return
        with open(self.json_file, 'w') as file:
            json.dump(self.to_json(), file, indent=4)

    # store a new version of the file in json
    # return the label of the last version stored
    def store_version(self, label: str, new_content: str) -> str:
        if not self.patches:
            self.patches.append(PatchInfo(label, new_content))
            return label
        
        dmp = diff_match_patch()
        last = self.patches[-1]
        del self.patches[-1]

        patch = dmp.patch_make(new_content, last.content) # type: ignore
        if not patch: # não houve mudança
            return last.label
        
        patch_text = dmp.patch_toText(patch) # type: ignore
        self.patches.append(PatchInfo(last.label, patch_text, len(last.content.splitlines())))
        self.patches.append(PatchInfo(label, new_content, len(new_content.splitlines())))
        return label

    def restore_all(self) -> list[PatchInfo]:
        output: list[PatchInfo] = []
        if not self.patches:
            return output
        last = self.patches[-1]
        output.append(PatchInfo(last.label, last.content))
        index = len(self.patches) -2
        dmp = diff_match_patch()
        while index > 0:
            label = self.patches[index].label
            patch = dmp.patch_fromText(self.patches[index].content) # type: ignore
            original = output[-1].content
            new_content = dmp.patch_apply(patch, original)[0] # type: ignore
            output.append(PatchInfo(label, new_content)) # type: ignore
            index -= 1
        output.reverse()
        return output

    def to_json(self):
        data: dict[str, list[dict[str, str]]] = {"patches": []}
        for patch in self.patches:
            data["patches"].append({"label": patch.label, "content": patch.content, "lines": str(patch.lines)})
        return data

    
    def __str__(self):
        return str(self.to_json())


from pathlib import Path

from tko.settings.rep_source import RepSource

from typing import Any

class RepData:
    def __init__(self):
        self.version: str = ""
        self.__sources: list[RepSource] = []
        self.expanded: list[str] = []
        self.tasks: dict[str, Any] = {}
        self.flags: dict[str, Any] = {}
        self.lang: str = ""
        self.selected: str = ""

    def set_source(self, source: RepSource):
        for i, s in enumerate(self.__sources):
            if s.name == source.name:
                self.__sources[i] = source
                return self
        self.__sources.append(source)

    def del_source(self, alias: str):
        self.__sources = [s for s in self.__sources if s.name != alias]
        return self

    def get_source(self, alias: str) -> RepSource | None:
        for s in self.__sources:
            if s.name == alias:
                return s
        return None

    def ensure_sandbox_source(self, rep_workspace: Path) -> None:
        sandbox_source = self.get_source(RepSource.STUDENT_SANDBOX_NAME)
        if sandbox_source is None:
            sandbox_source = RepSource("").set_student_sandbox().ensure_sandbox_source(rep_workspace)
            self.set_source(sandbox_source)
        else:
            sandbox_source.ensure_sandbox_source(rep_workspace)
        return None

    # fonte local é retornada primeiro para garantir que ela seja priorizada em relação a fontes externas
    def get_sources(self) -> list[RepSource]:
        external_sources: list[RepSource] = []
        local_sources: list[RepSource] = []
        for s in self.__sources:
            if s.is_sandbox_source():
                local_sources.append(s)
            else:
                external_sources.append(s)
        
        return local_sources + external_sources

    def get_expanded(self) -> list[str]:
        return self.expanded

    def get_tasks(self) -> dict[str, Any]:
        return self.tasks

    def get_flags(self) -> dict[str, Any]:
        return self.flags

    def get_lang(self) -> str:
        return self.lang

    def get_selected(self) -> str:
        return self.selected

    def set_lang(self, lang: str):
        self.lang = lang
        return self

    def _safe_load(self, data: dict[str, Any], key: str, target_type: type, default_value: Any = None):
        """Helper method to safely load a value from a dictionary."""
        if key in data and isinstance(data[key], target_type):
            return data[key]
        return default_value

    def load_from_dict(self, data: dict[str, Any]):
        try:
            # Load simple fields
            self.version = self._safe_load(data, "version", str, self.version)
            self.expanded = self._safe_load(data, "expanded", list, self.expanded)
            self.tasks = self._safe_load(data, "tasks", dict, self.tasks)
            self.flags = self._safe_load(data, "flags", dict, self.flags)
            self.lang = self._safe_load(data, "lang", str, self.lang)
            self.selected = self._safe_load(data, "selected", str, self.selected)
            # Load the 'source' field with specific validation
            if "sources" in data:
                source_data: list[dict[str, Any]] = data["sources"]
                if isinstance(source_data, list): # type: ignore
                    self.__sources = [RepSource("").load_from_dict(x) for x in source_data]
                else:
                    raise TypeError("The 'sources' field must be a list.")

        except (KeyError, TypeError) as e:
            print(f"Error loading data from dictionary: {e}")
            # Optionally, you can re-raise the exception or handle it differently
            # raise ValueError("Malformed data for RepData") from e

    def save_to_dict(self) -> dict[str, Any]:
        return {
            "version": self.version,
            "sources": [x.save_to_dict() for x in self.__sources],
            "expanded": self.expanded,
            "tasks": self.tasks,
            "flags": self.flags,
            "lang": self.lang,
            "selected": self.selected,
        }


class Colors:
    def __init__(self):
        self.focused_item: str = "B"
        self.task_text_done: str = "g"
        self.task_text_todo: str = "y"
        self.button_flag_on: str = "G"
        self.button_flag_off: str = "Y"
        self.progress_skill_done: str = "C"
        self.progress_skill_todo: str = "M"
        self.main_bar_done: str = "G"
        self.main_bar_todo: str = "R"
        self.task_skills: str = "c"
        self.task_new: str = "g"
        self.mark_nothing: str = "m"
        self.mark_started: str = "r"
        self.mark_required: str = "y"
        self.mark_complete: str = "g"
        self.filters: str = "Y"
    
    def to_dict(self):
        return self.__dict__
    
    def from_dict(self, attr_dict: dict[str, str]):
        for key, value in attr_dict.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self



import io
from pathlib import Path

from tko.logger.log_item_base import LogItemBase
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_move import LogItemMove
from tko.logger.log_item_self import LogItemSelf
from tko.play.patch_history import PatchHistory, PatchInfo
from tko.play.tracker import Tracker, Track
from tko.settings.rep_paths import RepPaths
from tko.util.decoder import Decoder
import datetime as dt
import csv
import enum
import os
from icecream import ic # type: ignore

class LogAction:
    def __init__(self, hash: str, timestamp: str, action: str, task: str, payload: str):
        self.hash = hash
        self.timestamp = timestamp
        self.action = action
        self.task = task
        self.payload = payload

class AType(enum.Enum):
    NONE = 'NONE'
    OPEN = 'OPEN' # Open program
    QUIT = 'QUIT' # Quit program
    DOWN = 'DOWN' # Down problem
    FREE = 'FREE' # Free Compile and RUN
    FAIL = 'FAIL' # Compile error or Execution error
    TEST = 'TEST' # run test and Coverage %
    SELF = 'SELF' # {c:{coverage}, a:{approach}, s:{autonomy}} after added {clear:%d, fun:%d, easy:%d}
    PICK = 'PICK' # Enter problem
    BACK = 'BACK' # Leave problem
    PROG = 'PROG' # deprecated
    SIZE = 'SIZE' # Problem files changes "{line_count:%d}"


class TrackerLoader: # deprecated
    @staticmethod
    def load_file_versions(task_track_folder: str) -> dict[str, dict[str, PatchInfo]]:
        files: list[str] = os.listdir(task_track_folder)
        files = [f for f in files if f.endswith(Tracker.extension)]
        file_versions: dict[str, dict[str, PatchInfo]] = {}
        for f in files:
            file_name = f[:-len(Tracker.extension)]
            file_path = os.path.join(task_track_folder, f)
            if not os.path.isfile(file_path):
                continue
            try:
                patch_history: list[PatchInfo] = PatchHistory().set_json_file(file_path).load_json().restore_all()
                file_versions[file_name] = {}
                for patch in patch_history:
                    file_versions[file_name][patch.label] = patch
            except Exception as e:
                ic(f"Error loading patch history from {file_path}: {e}")
        return file_versions

    @staticmethod
    def load_track_csv(task_track_folder: str) -> list[Track]:
        csv_file = os.path.join(task_track_folder, "track.csv")
        if not os.path.exists(csv_file):
            ic(f"CSV file {csv_file} does not exist.")
            return []
        return Tracker.load_from_log(csv_file)

    @staticmethod
    def load_from_task_track(task_track_folder: str) -> dict[dt.datetime, LogItemExec]:
        task = os.path.basename(task_track_folder)
        tracks: list[Track] = TrackerLoader.load_track_csv(task_track_folder)
        file_versions: dict[str, dict[str, PatchInfo]] = TrackerLoader.load_file_versions(task_track_folder)
        output: dict[dt.datetime, LogItemExec] = {}
        for track in tracks:
            track_datetime = Tracker.get_timestamp_from_string(track.timestamp)

            item = LogItemExec().set_datetime(track_datetime).set_key(task)
            rate = 0
            try:
                rate = int(track.result[:-1])  # Remove '%' and convert to int
            except ValueError:
                pass
            item.set_rate(rate)
            lines = 0
            for file_stamp in track.file_stamp_list:
                file, stamp = file_stamp.split(":")
                if file in file_versions:
                    if stamp in file_versions[file]:
                        patch_info = file_versions[file][stamp]
                        lines += patch_info.lines
            item.set_size(True, lines)
            output[item.get_datetime()] = item
        return output

class OldLogLoader:
    def __init__(self, rep_folder: Path):
        self.rep_folder: Path = rep_folder
        self.paths = RepPaths(rep_folder)
        self.base_dict: dict[dt.datetime, LogItemBase] = {}

        self.merge_old_log_into_base()
        self.merge_track_into_base()
    
    def merge_old_log_into_base(self):
        old_log_file = self.paths.get_old_history_file()
        entries: list[LogAction] = OldLogLoader.__load_file(old_log_file)
        for e in entries:
            item = OldLogLoader.__convert_to_base_list(e)
            if item is not None:
                self.base_dict[item.get_datetime()] = item

    def merge_track_into_base(self) -> None:
        track_exec_list: dict[dt.datetime, LogItemExec] = self.load_from_track_folder()
        for time, item_from_track in track_exec_list.items():
            done: bool = False
            if time in self.base_dict:
                item = self.base_dict[time]
                if isinstance(item, LogItemExec):
                    item_exec: LogItemExec = item
                    item_exec.set_size(True, item_from_track.get_size())
                    done = True
            if not done:
                item_exec = LogItemExec().set_datetime(time).set_key(item_from_track.get_key())
                item_exec.set_rate(item_from_track.get_rate())
                item_exec.set_size(True, item_from_track.get_size())
                self.base_dict[time] = item_exec        

    def load_from_track_folder(self) -> dict[dt.datetime, LogItemExec]:
        output: dict[dt.datetime, LogItemExec] = {}
        track_folder = self.paths.get_track_folder()
        if not os.path.exists(track_folder):
            return output
        entries = os.listdir(track_folder)
        for e in entries:
            folder_path = os.path.join(track_folder, e)
            if not os.path.isdir(folder_path):
                continue
            dict_stamp_exec: dict[dt.datetime, LogItemExec] = TrackerLoader.load_from_task_track(folder_path)
            output.update(dict_stamp_exec)

        return output
            

    @staticmethod
    def decode_self(e: LogAction) -> LogItemSelf:
        item = LogItemSelf().set_timestamp(e.timestamp).set_key(e.task)
        payload = e.payload.strip()

        if payload[0] == "{":
            payload = payload[1:-1]
            values = payload.split(",")
            kv: dict[str, str] = {}
            for svalue in values:
                k, v = svalue.split(":")
                kv[k.strip()] = v.strip()

            if "c" in kv:
                item.info.rate = int(kv["c"])
            if "cov" in kv:
                item.info.rate = int(kv["cov"])
            return item
        else:
            pass
        return item

    @staticmethod
    def __convert_to_base_list(e: LogAction) -> LogItemBase | None:

        item: LogItemBase | None = None
        if e.action == AType.PICK.value:
            item = LogItemMove().set_mode(LogItemMove.Mode.PICK).set_timestamp(e.timestamp).set_key(e.task)
        elif e.action == AType.BACK.value:
            item = LogItemMove().set_mode(LogItemMove.Mode.BACK).set_timestamp(e.timestamp).set_key(e.task)
        elif e.action == AType.DOWN.value:
            item = LogItemMove().set_mode(LogItemMove.Mode.DOWN).set_timestamp(e.timestamp).set_key(e.task)
        elif e.action == AType.TEST.value or e.action == AType.PROG.value:
            try:
                rate = int(e.payload)
            except ValueError:
                rate = 0
            item = LogItemExec().set_timestamp(e.timestamp).set_key(e.task)
            if rate != 0:
                item.set_rate(rate)
        elif e.action == AType.FREE.value:
            item = LogItemExec().set_mode(LogItemExec.Mode.FREE).set_timestamp(e.timestamp).set_key(e.task)
        elif e.action == AType.FAIL.value:
            item = LogItemExec().set_fail(LogItemExec.Fail.COMP).set_timestamp(e.timestamp).set_key(e.task)
        elif e.action == AType.SELF.value:
            item = OldLogLoader.decode_self(e)
        return item

    @staticmethod
    def __row_to_action_data( row: list[str]) -> LogAction | None:
        if len(row) < 5:
            return None
        hash = row[0]
        timestamp = row[1]
        action_value = row[2]
        task = row[3]
        payload = row[4]
        return LogAction(hash, timestamp, action_value, task, payload)

    @staticmethod
    def __load_file(history_file: Path | None) -> list[LogAction]:
        if history_file is None:
            return []
        if not os.path.exists(history_file):
            return []

        entries: list[LogAction] = []
        content = Decoder.load(history_file)
        with io.StringIO(content) as file:
            reader = csv.reader(file)
            rows = list(reader)
            for row in rows:
                action_data = OldLogLoader.__row_to_action_data(row)
                if action_data is not None:
                    entries.append(action_data)
        return entries



from __future__ import annotations

import time
from pathlib import Path
from datetime import  timedelta
import subprocess
import hashlib
from filelock import FileLock
import shutil


class GitCache:

    def __init__(self, cache_dir: str | Path, max_age: timedelta = timedelta(hours=1)) -> None:
        self.cache_dir: Path = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        self.max_age: timedelta = max_age

    def _repo_dir(self, url: str) -> Path:
        digest: str = hashlib.sha1(url.encode()).hexdigest()
        return self.cache_dir / digest

    def _lock_path(self, repo_path: Path) -> Path:
        return repo_path.with_suffix(".lock")

    def _git(self, *args: str, cwd: Path | None = None) -> None:
        subprocess.run(
            ["git", *args],
            cwd=cwd,
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            timeout=120,
        )

    def _clone(self, url: str, path: Path) -> None:
        self._git(
            "clone",
            "--depth", "1",
            "--filter=blob:none",
            "--no-single-branch",
            url,
            str(path),
        )

    def _update(self, repo: Path) -> None:
        self._git("fetch", "--prune", "origin", cwd=repo)
        self._git("reset", "--hard", "origin/HEAD", cwd=repo)
        self._git("clean", "-fd", cwd=repo)

    def _is_expired(self, repo: Path) -> bool:
        fetch_head: Path = repo / ".git" / "FETCH_HEAD"

        if not fetch_head.exists():
            return True

        age_seconds: float = time.time() - fetch_head.stat().st_mtime
        return age_seconds > self.max_age.total_seconds()

    def _acquire_lock(self, lock_path: Path):
        return FileLock(str(lock_path))

    def get(self, url: str, force_update: bool) -> Path:
        repo: Path = self._repo_dir(url)
        lock_path: Path = self._lock_path(repo)

        with self._acquire_lock(lock_path):

            if not repo.exists():
                print(f"Cloning {url} into cache...")
                self._clone(url, repo)
                return repo

            if self._is_expired(repo) or force_update:
                try:
                    print(f"Updating cache for {url}...")
                    self._update(repo)
                except subprocess.CalledProcessError:
                    print(f"Failed to update cache for {url}. Removing and re-cloning...")
                    shutil.rmtree(repo, ignore_errors=True)
                    self._clone(url, repo)

        return repo


from __future__ import annotations

import argparse
from pathlib import Path
import sys
from icecream import ic # type: ignore

from tko.feno.build import build_main
from tko.feno.indexer import indexer_main
from tko.feno.html import html_main
from tko.feno.mdpp import mdpp_main
from tko.feno.older import older_main
from tko.feno.remote_md import remote_main
from tko.feno.filter import filter_main, build_drafts_main

from tko.cmds.cmd_task import CmdTask
from tko.cmds.cmd_collect import CmdCollect
from tko.cmds.cmd_open import CmdOpen
from tko.cmds.cmd_run import Run
from tko.cmds.cmd_build import CmdBuild
from tko.cmds.cmd_config import CmdConfig, ConfigParams
from tko.enums.diff_mode import DiffMode
from tko.util.text import AnsiColor
from tko.enums.diff_count import DiffCount
from tko.mico.pull import pull_all_parallel_main

from tko.util.param import Param
from tko.util.pattern_loader import PatternLoader

from tko.settings.settings import Settings
from tko.cmds.cmd_diff import cmd_diff

from tko.settings.repository import Repository
from tko.settings.rep_paths import RepPaths
from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols
from tko.settings.check_version import CheckVersion
from tko.settings.rep_starter import RepStarter
from tko.settings.rep_source_actions import RepSourceActions

from tko.__init__ import __version__


class Main:

    @staticmethod
    def run(args: argparse.Namespace) -> None:
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        if args.none:
            param.set_diff_count(DiffCount.NONE)
        elif args.all:
            param.set_diff_count(DiffCount.ALL)
        else:
            param.set_diff_count(DiffCount.FIRST)

        if args.filter:
            param.set_filter(True)
        if args.compact:
            param.set_compact(True)

        settings = Settings()
        if not args.side and not args.down:
            param.set_diff_mode(settings.app.get_diff_mode())
        elif args.side:
            param.set_diff_mode(DiffMode.SIDE)
        elif args.down:
            param.set_diff_mode(DiffMode.DOWN)
        cmd_run = Run(settings, args.target_list, param)
        if args.eval:
            cmd_run.show_track_info().show_self_info()

        cmd_run.execute()

    @staticmethod
    def tui(args: argparse.Namespace) -> None:
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        settings = Settings()
        param.set_diff_mode(settings.app.get_diff_mode())
        if args.filter:
            param.set_filter(True)
        cmd_run = Run(settings, args.target_list, param)
        cmd_run.set_curses()
        cmd_run.execute()

    @staticmethod
    def build(args: argparse.Namespace):
        settings = Settings()
        PatternLoader.pattern = args.pattern
        manip = Param.Manip().set_unlabel(args.unlabel).set_to_sort(args.sort).set_to_number(args.number)
        build = CmdBuild(Path(args.target), [Path(x) for x in args.target_list], manip, settings)
        build.execute()

    @staticmethod
    def diff(args: argparse.Namespace):
        cmd_diff(args)

    @staticmethod
    def open(args: argparse.Namespace):
        settings = Settings()
        update: bool = args.update
 
        if args.global_cache:
            RepPaths.use_global_cache_folder = True
            print(f"Usando cache global em: {RepPaths('').get_cache_folder()}")

        changedir = args.changedir

        rec_folder = RepPaths.rec_search_for_repo_parents(changedir)
        if rec_folder is not None:
            changedir = rec_folder
        action = CmdOpen(settings, update).load_folder(changedir)
        if not CheckVersion(settings).is_updated():
            action.display_need_update()
        action.execute()

    @staticmethod
    def collect_task(args: argparse.Namespace):
        settings = Settings()
        changedir = args.changedir
        rec_folder = RepPaths.rec_search_for_repo_parents(changedir)
        if rec_folder is not None:
            changedir = rec_folder
        rep = Repository(changedir)
        width: int  = args.graph[0]
        height: int  = args.graph[1]
        CmdTask.show_graph(settings, rep, args.label, width, height)

    @staticmethod
    def init(args: argparse.Namespace):
        changedir: Path | None = args.changedir
        language: str | None = args.language
        rep_starter = RepStarter(folder=changedir, language=language)
        rep_starter.execute()

    @staticmethod
    def remote_ls(args: argparse.Namespace):
        changedir: Path | None = args.changedir
        try:
            rep_actions = RepSourceActions(changedir)
            rep_actions.remote_ls()
        except ValueError as e:
            print(f"Erro ao listar fontes: {e}")

    @staticmethod
    def remote_rm(args: argparse.Namespace):
        name: str = args.name
        changedir: Path | None = args.changedir
        try:
            rep_actions = RepSourceActions(changedir)
            rep_actions.remote_rm(alias=name)
        except ValueError as e:
            print(f"Erro ao deletar fonte: {e}")

    @staticmethod
    def remote_add(args: argparse.Namespace):
        target: str = args.target
        default_git_alias: str | None = None
        git_repository_url: str | None = None
        local_source_dir: str | None = None

        if target.startswith("@"):
            default_git_alias = target[1:]
        elif target.startswith("http:") or target.startswith("https:") or target.startswith("ssh:"):
            git_repository_url = target
        else:
            local_source_dir = target

        git_branch: str = args.branch
        writeable: bool = args.write

        changedir: Path | None = args.changedir
        name: str = args.name
        quest_filter: list[str] | None = args.quest
        task_filter: list[str] | None = args.task
        try:
            rep_actions = RepSourceActions(changedir)
            rep_actions.remote_add(
                name=name, 
                remote_default=default_git_alias, 
                branch=git_branch, 
                remote_url=git_repository_url, 
                remote_dir=local_source_dir, 
                quest_filter=quest_filter, 
                task_filter=task_filter,
                writeable=writeable)
            rep_actions.print_end_msg()
        except ValueError as e:
            print(f"Erro ao adicionar fonte: {e}")

    @staticmethod
    def remote_filter(args: argparse.Namespace):
        name: str = args.name
        changedir: Path | None = args.changedir
        quests: list[str] | None = args.quest
        tasks: list[str] | None = args.task
        clear: bool = args.clear
        if clear and (quests or tasks):
            print("Erro: --clear não pode ser usado com --quest ou --task")
            return
        rep_actions = RepSourceActions(changedir)
        rep_actions.remote_filter(alias=name, quests=quests, tasks=tasks, clear=clear)

    @staticmethod
    def config(args: argparse.Namespace):
        settings = Settings()
        param = ConfigParams()
        param.side = args.side
        param.down = args.down
        param.images = args.images
        param.editor = args.editor
        param.borders = args.borders
        param.timeout = args.timeout
        CmdConfig.execute(settings, param)

    @staticmethod
    def list(_: argparse.Namespace):
        settings = Settings()
        print(str(settings))

class Parser:
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(prog='tko',
                                                                       description=f'tko {__version__}', add_help=False)
        
        self.parser.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        self.subparsers = self.parser.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')

        self.parent_manip = Parser.create_parent_manip()
        self.parent_basic = Parser.create_parent_basic()

        self.add_parser_global()
        self.create_parent_basic()
        self.create_parent_manip()
        self.add_parser_rep_actions()
        self.add_parser_task()
        self.add_parser_config()
        self.add_parser_collect()
        self.add_parser_run()
        self.add_parser_build()
        self.add_parser_class()

    def add_parser_global(self):
        self.parser.add_argument('-c', '--configfile', metavar='FILE', type=str, help='Global config file')
        self.parser.add_argument('-w', '--width', metavar='WIDTH', type=int, help="Terminal width")
        self.parser.add_argument('-v', "--version", action='store_true', help='Show version and exit')
        self.parser.add_argument('-m', "--mono", action='store_true', help='Disable colors')
        self.parser.add_argument('-D', '--debug', action='store_true', help='Enable debug mode')
        self.parser.add_argument('-C', '--changedir', metavar='DIR', type=Path, default=Path('.'), help='Repository directory')

    @staticmethod
    def create_parent_basic():
        parent_basic = argparse.ArgumentParser(add_help=False)
        parent_basic.add_argument('-i', '--index', metavar="I", type=int, help='Run a specific test index')
        parent_basic.add_argument('-p', '--pattern', metavar="P", type=str, default='@.in @.sol',
                                  help='Input/output file pattern (default: "@.in @.sol")')
        return parent_basic

    @staticmethod
    def create_parent_manip():
        parent_manip = argparse.ArgumentParser(add_help=False)
        parent_manip.add_argument('--width', '-w', type=int, help="Terminal width")
        parent_manip.add_argument('--unlabel', '-u', action='store_true', help='Remove all labels')
        parent_manip.add_argument('--number', '-n', action='store_true', help='Number labels')
        parent_manip.add_argument('--sort', '-s', action='store_true', help="Sort test cases by input size")
        parent_manip.add_argument('--pattern', '-p', metavar="@.in @.out", type=str, default='@.in @.sol',
                                  help='Input/output file pattern (default: "@.in @.sol")')
        return parent_manip

    def add_parser_run(self):
        parser_run = self.subparsers.add_parser('run', help='Runs a task in raw terminal', parents=[self.parent_basic], add_help=False)
        parser_run.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        parser_run.add_argument('target_list', metavar='TARGET', type=str, nargs='*', help='Solvers files, test cases or directories containing them')
        parser_run.add_argument('-f', '--filter', action='store_true', help='Filter solver files in temporary directory before running')
        parser_run.add_argument('-e', '--eval', action='store_true', help='Show percentage of passed tests')
        parser_run.add_argument('-c', '--compact', action='store_true', help='Hide test case descriptions in failures')

        group_n = parser_run.add_mutually_exclusive_group()
        group_n.add_argument('-n', '--none', action='store_true', help='Hide all failures')
        group_n.add_argument('-a', '--all', action='store_true', help='Show all failures')

        # add an exclusive group for diff mode
        group = parser_run.add_mutually_exclusive_group()
        group.add_argument('-d', '--down', action="store_true", help="Diff mode: top-to-bottom")
        group.add_argument('-s', '--side', action="store_true", help="Diff mode: side-by-side")
        parser_run.set_defaults(func=Main.run)

    def add_parser_task(self):
        parser_task = self.subparsers.add_parser('task', help='Manage individual tasks', add_help=False)
        parser_task.add_argument( "-h", "--help", action="help", help="Show help message and exit" )

        subpar_task = parser_task.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')
        parser_open = subpar_task.add_parser('open', parents=[self.parent_basic], help='Open a task in tui')
        parser_open.add_argument("-f", "--filter", action='store_true', help='Filter solver files in temporary directory before running')
        parser_open.add_argument('target_list', metavar='T', type=str, nargs='*', help='Solvers, test cases or directories to load')
        parser_open.set_defaults(func=Main.tui)

    def add_parser_config(self):
        parser_cfg = self.subparsers.add_parser('config', help='DESCRIPTION', add_help=False)
        parser_cfg.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        subpar_repo = parser_cfg.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION') 

        cfg_reset = subpar_repo.add_parser('reset',
                                           help='Reset configuration to factory default')
        cfg_reset.set_defaults(func=CmdCollect.reset)

        cfg_set = subpar_repo.add_parser('set', help='Set default configuration values')

        g_diff = cfg_set.add_mutually_exclusive_group()
        g_diff.add_argument('--side', action='store_true', help='Set side_by_side diff mode')
        g_diff.add_argument('--down', action='store_true', help='Set up_to_down   diff mode')

        cfg_set.add_argument("--editor", metavar="cmd", type=str, help='Set editor command')
        cfg_set.add_argument("--borders", metavar="[0|1]", type=str, help='Enable borders')
        cfg_set.add_argument("--images", metavar="[0|1]", type=str, help='Enable images')
        cfg_set.add_argument("--timeout", metavar="sec", type=int, help='Set timeout')

        cfg_set.set_defaults(func=Main.config)

        cfg_list = subpar_repo.add_parser('list', help='List default configuration values')
        cfg_list.set_defaults(func=Main.list)

    def add_parser_class(self):
        parser_class = self.subparsers.add_parser('class', help='Manage class tasks', add_help=False)
        parser_class.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        subpar_class = parser_class.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')

        collect_cmd = subpar_class.add_parser("collect", help="Colect and merge data from many repos")
        collect_cmd.add_argument("path", type=str, nargs="+", help="Paths to repos")
        collect_cmd.add_argument("--json", "-j", type=str, help="Path to save the extracted JSON data")
        collect_cmd.add_argument("--csv", "-c", type=str, help="Path to save the extracted CSV data")
        collect_cmd.add_argument("--block_prefix", "-b", type=str, help="Block prefix to insert in csv file")
        collect_cmd.set_defaults(func=CmdCollect.collect_batch)

        pull_cmd = subpar_class.add_parser("pull", help="Perform git pull in many repos using threads")
        pull_cmd.add_argument("path", type=str, nargs="+", help="Paths to repos")
        pull_cmd.add_argument("-t", "--threads", type=int)
        pull_cmd.set_defaults(func=pull_all_parallel_main)

    def add_parser_collect(self):
        parser_repo = self.subparsers.add_parser('collect', help='Collect evaluation data', add_help=False)
        parser_repo.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        subpar_collect = parser_repo.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')

        # repo_update = subpar_repo.add_parser("update", parents=[self.parent_folder], help="Update repository cache")
        # repo_update.set_defaults(func=CmdRep.update)

        parser_task = subpar_collect.add_parser('task', help='Collect data from task', add_help=False)
        parser_task.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        parser_task.add_argument('label', type=str, help='Task label')
        parser_task.add_argument('graph', nargs=2, type=int, metavar=('WIDTH', 'HEIGHT'), help="Show task graph")
        parser_task.set_defaults(func=Main.collect_task)

        repo_collect = subpar_collect.add_parser("repo", help="Collect data from this repository")
        repo_collect.add_argument("--json", action='store_true', help="Collect as json data")
        repo_collect.add_argument("--resume", action='store_true', help="Collect resume")
        repo_collect.add_argument("--log", action='store_true', help="Collect history log")
        repo_collect.add_argument('--game', action='store_true', help="Collect game info")
        repo_collect.add_argument('--daily', action='store_true', help="Daily graph")
        repo_collect.add_argument('--width', type=int, default=100, help="Daily graph width")
        repo_collect.add_argument('--height', type=int, default=10, help="Daily graph height")
        repo_collect.add_argument('--color', type=int, default=1, help="Daily graph color [0|1]")
        repo_collect.set_defaults(func=CmdCollect.collect_main)


    def add_parser_rep_actions(self):
        parser_init = self.subparsers.add_parser('init', help='Initialize empty TKO repository', add_help=False)
        parser_init.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        parser_init.add_argument('-l', '--language', type=str, metavar=('LANG'), help='Default repository language (e.g. py, cpp, java, go)')
        parser_init.set_defaults(func=Main.init)

        parser_open = self.subparsers.add_parser('open', help='Open repository in interactive mode', add_help=False)
        parser_open.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        parser_open.add_argument('-g', '--global-cache', action='store_true', help='Use global cache for remote URL sources')
        parser_open.add_argument('-u', '--update', action='store_true', help='Force update remote URL sources')
        parser_open.set_defaults(func=Main.open)

        parser_source = self.subparsers.add_parser('remote', help='Manage remote task sources', add_help=False)
        parser_source.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        sub_source = parser_source.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')
        # create subcommands inside source
        source_add = sub_source.add_parser("add", help="Add a new task source", add_help=False)
        source_add.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        source_add.add_argument('name', type=str, help='Name of the remote')
        source_add.add_argument('target', type=str, metavar=('TARGET'), help='Remote source: git URL, local directory or preset name (e.g. +fup, +ed, +poo)')
        source_add.add_argument('-q', '--quest', action='append', metavar=('QUEST_ID'), type=str, help='Load tasks only from selected quests')
        source_add.add_argument('-t', '--task', action='append', metavar=('TASK_ID'), type=str, help='Load tasks only from selected tasks')
        
        source_add.add_argument('-b', '--branch', type=str, default='master', help='Branch name for git remote sources')
        source_add.add_argument('-w', '--write', action='store_true', help='Allow modifications for local directory remotes (default: readonly)')
        source_add.set_defaults(func=Main.remote_add)

        source_list = sub_source.add_parser("ls", help="List remote task sources", add_help=False)
        source_list.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        source_list.set_defaults(func=Main.remote_ls)

        source_del = sub_source.add_parser("rm", help="Remove a remote task source", add_help=False)
        source_del.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        source_del.add_argument('name', type=str, help='Name of the remote to be removed')
        source_del.set_defaults(func=Main.remote_rm)

        source_filter = sub_source.add_parser("filter", help="Manage filters for a remote task source", add_help=False)
        source_filter.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        source_filter.add_argument('name', type=str, help='Name of the remote')
        source_filter.add_argument('-q', '--quest', action='append', metavar=('QUEST_ID'), type=str, help='Add quests to the filter list')
        source_filter.add_argument('-t', '--task', action='append', metavar=('TASK_ID'), type=str, help='Add tasks to the filter list')
        source_filter.add_argument('--clear', action='store_true', help='Clear all filters')
        source_filter.set_defaults(func=Main.remote_filter)
        

    def add_parser_build(self):
        parser_build = self.subparsers.add_parser('build', help='Build repository artifacts', add_help=False)
        parser_build.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        subparsers = parser_build.add_subparsers(title='subcommands', metavar='COMMAND', help='DESCRIPTION')

        parser_b = subparsers.add_parser('tests', parents=[self.parent_manip], help='Build a test target', add_help=False)
        parser_b.add_argument( "-h", "--help", action="help", help="Show help message and exit" )
        parser_b.add_argument('target', metavar='T_OUT', type=str, help='Target to be built')
        parser_b.add_argument('target_list', metavar='T', type=str, nargs='+', help='Input test targets')
        parser_b.set_defaults(func=Main.build)

        # subparser for the 'build' command
        parser_r = subparsers.add_parser('mapi', help='Build .mapi file', add_help=False)
        parser_r.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_r.add_argument('targets', metavar='T', type=str, nargs='*', help='directories')
        parser_r.add_argument("--check", "-c", action="store_true", help="Check if the file needs to be rebuilt")
        parser_r.add_argument("--brief", "-b", action="store_true", help="Brief mode")
        # parser_r.add_argument("--pandoc", "-p", action="store_true", help="Use pandoc rather than python markdown")
        parser_r.add_argument("--remote", "-r", action="store_true", help="Search for remote.cfg file and create absolute links")
        parser_r.add_argument("--erase", "-e", action="store_true", help="Erase .md and .tio temp files")
        parser_r.add_argument("--debug", "-d", action='store_true', help="Display debug msgs")
        parser_r.set_defaults(func=build_main)

        # subparser for the 'indexer' command
        parser_i = subparsers.add_parser('indexer', help='Index Readme file', add_help=False)
        parser_i.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_i.add_argument('path', type=str, help='Path to Markdown file')
        parser_i.add_argument("base", type=str, help="Folder with the problems")
        parser_i.set_defaults(func=indexer_main)

        # subparser for the 'html' command
        parser_h = subparsers.add_parser('html', help='Generate HTML file from markdown file', add_help=False)
        parser_h.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_h.add_argument('input', type=str, help='Input markdown file')
        parser_h.add_argument('output', type=str, help='Output HTML file')
        parser_h.add_argument('--title', type=str, default="Problema", help='Title of the HTML file')
        parser_h.set_defaults(func=html_main)

        # subparser for the 'mdpp' command
        parser_m = subparsers.add_parser('mdpp', help='Preprocessor for markdown files', add_help=False)
        parser_m.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_m.add_argument('targets', metavar='T', type=str, nargs='*', help='Readme files or None to default task behavior')
        parser_m.add_argument('--quiet', '-q', action="store_true", help='quiet mode')
        parser_m.add_argument('--clean', '-c', action="store_true", help='clean mode')
        parser_m.set_defaults(func=mdpp_main)

        # subparser for the 'older' command
        parser_o = subparsers.add_parser('older', help='Check if the source is newer than the target', add_help=False)
        parser_o.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_o.add_argument('targets', type=str, nargs='+', help='Target files or directories')
        parser_o.set_defaults(func=older_main)

        parser_diff = subparsers.add_parser('diff', help='Show diff for 2 inputs or files', add_help=False)
        parser_diff.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        exclusive_group_target = parser_diff.add_mutually_exclusive_group(required=True)
        exclusive_group_target.add_argument('--path', '-f', action='store_true', help='Targets are paths')
        exclusive_group_target.add_argument('--text', '-t', action='store_true', help='Compare two texts')
        parser_diff.add_argument('target_a', type=str, help='First target to be compared')
        parser_diff.add_argument('target_b', type=str, help='Second target to be compared')
        exclusive_group = parser_diff.add_mutually_exclusive_group()
        exclusive_group.add_argument('--side', '-s', action='store_true', help="Diff mode side-by-side")
        exclusive_group.add_argument('--down', '-d', action='store_true', help="Diff mode up-to-down")
        parser_diff.set_defaults(func=Main.diff)

        # subparser for the 'redirect' command
        parser_rm = subparsers.add_parser('redirect', help='Create redirected markdown file', add_help=False)
        parser_rm.add_argument( "-h", "--help", action="help", help="Show help message and exit")
        parser_rm.add_argument('target', type=str, help='Directories')
        parser_rm.add_argument('--output', '-o', type=str, help='Output file')
        parser_rm.set_defaults(func=remote_main)

        # subparser for the 'filter' command
        parser_f = subparsers.add_parser('filter', help='Filter code removing answers', add_help=False)
        parser_f.add_argument( "-h", "--help", action="help", help="Show help message and exit")

        parser_f.add_argument('target', type=str, help='file or directory to process')
        parser_f.add_argument('-u', '--update', action="store_true", help='update source file')
        parser_f.add_argument('-c', '--cheat', action="store_true", help='recursive cheat mode cleaning comments on students files')
        parser_f.add_argument('-o', '--output', type=str, help='output target')
        parser_f.add_argument("-r", "--recursive", action="store_true", help="recursive mode")
        parser_f.add_argument("-f", "--force", action="store_true", help="force mode")
        parser_f.add_argument("-q", "--quiet", action="store_true", help="quiet mode")
        parser_f.add_argument("-i", "--indent", type=int, default=0, help="indent using spaces")
        parser_f.set_defaults(func=filter_main)

        parser_drafts = subparsers.add_parser('drafts', help='Create drafts for TKO task using src dir', add_help=False)
        parser_drafts.add_argument("-h", "--help", action="help", help="Show help message and exit")
        parser_drafts.set_defaults(func=build_drafts_main)


def execute(parser: argparse.ArgumentParser, args: argparse.Namespace) -> int:
    settings = Settings()
    if args.width is not None:
        RawTerminal.set_terminal_size(args.width)
    if args.configfile:
        settings.set_settings_file(args.configfile)
    settings.load_settings()

    if args.mono:
        AnsiColor.enabled = False
    else:
        AnsiColor.enabled = True
        symbols.set_colors()

    if args.debug:
        ic.configureOutput(includeContext=True, outputFunction=print)
        ic.enable()
        ic("Debug mode enabled")
    else:
        ic.disable()

    if args.version:
        if args.version:
            print("tko " + __version__)
            return 0
    else:
        if "func" in args:
            code = args.func(args)
            return code if code is not None else 0
        else:
            parser.print_help()
    return 0


def main():
    try:
        parser = Parser().parser
        args = parser.parse_args()
        execute(parser, args)
        sys.exit(0)
    except KeyboardInterrupt:
        print("\n\nKeyboard Interrupt")
        sys.exit(1)
    except Warning as w:
        print(w)
        sys.exit(1)
    # except Exception as e:
    #     print(e)
    #     sys.exit(1)

if __name__ == '__main__':
    main()



from tko.play.floating_drop_down import FloatingDropDown
from tko.play.flags import FlagsMan
from tko.play.floating import Floating
from tko.play.floating_drop_down import FloatingInputData
from tko.play.floating_manager import FloatingManager
from tko.util.text import Text
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.down.drafts import Drafts


class LanguageSetter:

    @staticmethod
    def check_lang_in_text_mode(repo: Repository):
        lang = repo.data.lang
        if lang == "":
            options = Drafts().get_languages_with_drafts()
            print("\nLinguagem padrão ainda não foi definida.\n")
            while True:
                print("Escolha entre as opções a seguir ", end="")
                print("[" + ", ".join(options) + "]", ":", end=" ")
                lang = input()
                if lang in options:
                    break
            repo.data.lang = lang
            repo.save_config()
            
    def __init__(self, settings: Settings, rep: Repository, flagsman: FlagsMan, fman: FloatingManager):
        self.flagsman = flagsman
        self.fman = fman
        self.rep = rep
        self.settings = settings

    def set_language(self):
        options: list[FloatingInputData] = []
        for lang in Drafts().get_languages_with_drafts():
            options.append(FloatingInputData(TextFunctor(lang), SetLangFunctor(self.settings, self.rep, self.fman, lang)))

        self.fman.add_input(
            FloatingDropDown().set_floating(
                Floating()
                .top()
                .set_header(" Escolha a extensão default para os rascunhos ")
                .set_footer(" Escolha e reinicie o tko para aplicar!!!!! ")
                .set_text_ljust()
            )
            .set_options(options)
            .set_default_index(Drafts().get_languages_with_drafts().index(self.rep.data.get_lang()))
        )

class TextFunctor:
    def __init__(self, data: str):
        self.value = data
    
    def __call__(self):
        return Text().add(self.value)

class SetLangFunctor:
    def __init__(self, settings: Settings, rep: Repository, fman: FloatingManager, lang: str):
        self.rep = rep
        self.fman = fman
        self.value = lang
        self.settings = settings

    def __call__(self):
        self.rep.data.set_lang(self.value.strip())
        self.rep.save_config()
        self.fman.add_input(
            Floating()
            .bottom()
            .right()
            .put_text("")
            .put_text("Linguagem alterada para " + self.value)
            .put_text("")
            .set_warning()
        )



from __future__ import annotations
from tko.game.task import Task

import os
import re
from icecream import ic # type: ignore
from pathlib import Path

class TaskParser:

    def __init__(self, index_path: Path, source_alias: str):
        self.index_path = index_path
        self.task: Task | None = Task().set_remote_name(source_alias)
            
    @staticmethod
    def filter_task_key(key: str) -> str:
        allowed = "0123456789_abcdefghijklmnopqrstuvwxyz-ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        new_key = ""
        for c in key:
            if c in allowed:
                new_key += c
            else:
                break
        return new_key

    # returns
    # tuple[bool, title, html_tags, link]
    def match_full_pattern(self, line: str) -> tuple[bool, str, str, str]:
        pattern = r'\s*?- \[ \](.*?)\[([^\]]+)\]\(([^)]+)\)?'
        match = re.match(pattern, line)
        if match is None:
            return False, "", "", ""
        tags = match.group(1).strip()
        tags = tags.replace("`", " ").replace("<!--", " ").replace("-->", " ")
        title = match.group(2).strip()
        link = match.group(3).strip()
        return True, tags, title, link

    def decode_task_types(self, info: str):
        if self.task is None:
            return
        for c in info:
            # if c is digit, set xp
            if c.isdigit():
                self.task.xp = int(c)
            elif c == "a":
                self.task.task_rate = Task.TaskRate.AUTO
            elif c == "u":
                self.task.task_rate = Task.TaskRate.USER
            elif c == "t":
                self.task.task_rate = Task.TaskRate.TICK
            elif c == "=":
                self.task.task_path = Task.TaskPath.MAIN
            elif c == "+":
                self.task.task_path = Task.TaskPath.SIDE
            elif c == "?":
                self.task.task_help = Task.TaskHelp.OPEN
            elif c == "!":
                self.task.task_help = Task.TaskHelp.EXAM
            elif c == ">":
                self.task.task_action = Task.TaskAction.VIEW
            elif c == "#":
                self.task.task_action = Task.TaskAction.EDIT
            else:
                raise Warning(f"Parsing {self.index_path}:{self.task.line_number}, Tipo de tarefa desconhecido: {c}")

    def __parse_key_task_types(self, tags: str) -> str:
        if self.task is None:
            return ""
        new_title: list[str] = []
        words = [w for w in tags.split(" ") if w != ""]
        for item in words:
            if item.startswith("@"):
                self.task.set_key(self.filter_task_key(item[1:]))
            elif item.startswith(":"):
                self.decode_task_types(item[1:])
            else:
                new_title.append(item)
        return " ".join(new_title)

    def redirect_from_readme(self, link: str) -> str:
        if not os.path.isabs(link):
            basedir = os.path.dirname(self.index_path)
            target = os.path.join(basedir, link)
            return target
        return link

    def parse_line(self, line: str, line_num: int = 0) -> TaskParser:
        found, tags, title, link = self.match_full_pattern(line)
        if not found:
            self.task = None
        if self.task is None:
            return self
        
        task = self.task
        task.line_number = line_num
        task.line = line
        title = self.__parse_key_task_types(tags + " " + title)
        task.set_title(title)

        if task.get_key() == "":
            self.task = None
            return self

        if link.startswith("http://") or link.startswith("https://"):
            self.task.set_remote_view_type()
            self.task.target = link
            return self
        
        task.set_origin_folder(Path(os.path.dirname(self.redirect_from_readme(link))))
        if task.task_action == Task.TaskAction.VIEW:
            self.task.target = self.redirect_from_readme(link)
        else:
            self.task.target = link

        return self

    def get_task(self) -> Task | None:
        return self.task


    def check_path_try(self):
        if self.task is None:
            return self
        if self.task.is_import_type():
            relative_path = self.index_path.parent / self.task.target
            if not relative_path.exists():
                raise Warning(f"Parsing {self.index_path}, Arquivo de tarefa não encontrado: {self.task.target}")
        return self


from tko.util.text import Text
from tko.play.floating import Floating, FloatingABC
from tko.settings.settings import Settings


class FloatingCalibrate(FloatingABC):
    def __init__(self, settings: Settings):
        self.floating = Floating()
        self.settings = settings
        self.floating.set_header(" Calibrar teclas direcionais ")
        # self.set_text_ljust()
        self.floating.set_footer_text(Text.format(" Use {y:Enter} para salvar, {y:q} para cancelar e {y:r} para reiniciar. "))
        self._index = 0
        self._options: list[int] = [settings.app.get_key_left(),
                                   settings.app.get_key_right(),
                                   settings.app.get_key_up(),
                                   settings.app.get_key_down(), 
                                   settings.app.get_key_esc(),
                                   settings.app.get_key_pg_up(),
                                   settings.app.get_key_pg_down(), 
                                   settings.app.get_key_backspace()]
        self.floating.frame.set_border_color("m")
        self._exit_on_action = True
        self.right_dx = 5 # shortcut space
        self.settings = settings
        self.set_key_content()

    def draw(self):
        self.floating.draw()

    def is_enable(self) -> bool:
        return self.floating.is_enable()

    def set_key_content(self):
        def format_value(value: int) -> str:
            if value == 0:
                return "---"
            return f"{value:>3}"

        content = self.floating.content
        content.clear()
        color = "G" if self._index == 0 else ""
        content.append(Text().addf(color, "Left      ").addf(color, format_value(self._options[0])))
        color = "G" if self._index == 1 else ""
        content.append(Text().addf(color, "Right     ").addf(color, format_value(self._options[1])))
        color = "G" if self._index == 2 else ""
        content.append(Text().addf(color, "Up        ").addf(color, format_value(self._options[2])))
        color = "G" if self._index == 3 else ""
        content.append(Text().addf(color, "Down      ").addf(color, format_value(self._options[3])))
        color = "G" if self._index == 4 else ""
        content.append(Text().addf(color, "Esc       ").addf(color, format_value(self._options[4])))
        color = "G" if self._index == 5 else ""
        content.append(Text().addf(color, "PageUp    ").addf(color, format_value(self._options[5])))
        color = "G" if self._index == 6 else ""
        content.append(Text().addf(color, "PageDown  ").addf(color, format_value(self._options[6])))
        color = "G" if self._index == 7 else ""
        content.append(Text().addf(color, "Backspace ").addf(color, format_value(self._options[7])))

    def process_input(self, key: int) -> int:
        # self.draw()
        
        if key == ord('\n'):
            self.floating.enable = False
            self.settings.app.set_key_left(self._options[0])
            self.settings.app.set_key_right(self._options[1])
            self.settings.app.set_key_up(self._options[2])
            self.settings.app.set_key_down(self._options[3])
            self.settings.app.set_key_esc(self._options[4])
            self.settings.app.set_key_pg_up(self._options[5])
            self.settings.app.set_key_pg_down(self._options[6])
            self.settings.app.set_key_backspace(self._options[7])
            self.settings.save_settings()
            return -1
        elif key == ord('q'):
            self.floating.enable = False
            return -1
        elif key == ord('r'):
            self._index = 0
            self.set_key_content()
        elif not (ord('a') <= key <= ord('z') or ord('A') <= key <= ord('Z')) or ord('0') <= key <= ord('9'):
            self._options[self._index] = key
            for i in range(len(self._options)):
                if i != self._index and self._options[i] == key:
                    self._options[i] = 0
            self._index += 1
            if self._index >= len(self._options):
                self._index = 0
            self.set_key_content()
        return -1



from __future__ import annotations
from typing import Any
import re
import os
import tomllib

from tko.run.unit import Unit
from tko.util.pattern_loader import PatternLoader
from tko.util.decoder import Decoder
from pathlib import Path


class CaseData:
    def __init__(self, case: str="", inp: str="", outp: str="", grade: None | int = None):
        self.case: str = case
        self.input: str = VplParser.finish(inp)
        self.output: str = VplParser.unwrap(VplParser.finish(outp))
        self.grade: None | int = grade

    # @override
    def __str__(self) -> str:
        return "case=" + self.case + '\n' \
                + "input=" + self.input \
                + "output=" + self.output \
                + "gr=" + str(self.grade)

class VplParser:
    @staticmethod
    def finish(text: str):
        return text if text.endswith("\n") else text + "\n"

    @staticmethod
    def unwrap(text: str):
        while text.endswith("\n"):
            text = text[:-1]
        if text.startswith("\"") and text.endswith("\""):
            text = text[1:-1]
        return VplParser.finish(text)

    regex_vpl_basic = r"[cC]ase *= *([ \S]*) *\n *[iI]nput *=([\s\S]*?)^ *[oO]utput *=([\s\S]*?)(?=^ *[cC]ase *=|\Z)"
    regex_grade_reduction = r"([\s\S]*) *[Gg]rade [rR]eduction *= *(.*)%"

    @staticmethod
    def filter_quotes(x: str):
        return x[1:-2] if x.startswith('"') else x

    @staticmethod
    def parse_vpl(content: str) -> list[CaseData]:
        output: list[CaseData] = []
        for m in re.finditer(VplParser.regex_vpl_basic, content, re.MULTILINE):
            str_case = m.group(1)
            str_input = m.group(2)
            str_output = m.group(3)
            grade: int | None = None
            gr = re.match(VplParser.regex_grade_reduction, str_output, re.MULTILINE)
            if gr is not None:
                str_output = gr.group(1)
                grade = int(gr.group(2))
            output.append(CaseData(str_case, str_input, str_output, grade))
        return output

    @staticmethod
    def to_vpl(unit: CaseData):
        text = "case=" + unit.case + "\n"
        text += "inserted=" + unit.input
        text += "expected=\"" + unit.output + "\"\n"
        if unit.grade is not None:
            text += "grade reduction=" + str(unit.grade) + "%\n"
        return text

def parse_toml(content: str) -> list[dict[str, Any]]:
    main_list_name = "tests"

    try:
        data = tomllib.loads(content)
    except tomllib.TOMLDecodeError as _:
        return []

    if main_list_name not in data or not isinstance(data[main_list_name], list):
        return []

    result: list[dict[str, Any]] = []

    for i, case in enumerate(data[main_list_name], start=1): # type:ignore
        if not isinstance(case, dict):
            raise ValueError(f"Case {i} inválido.")

        if "input" not in case or "output" not in case:
            print(f"warning: case {i} data: {case}") # type: ignore
            continue # ou raise ValueError(f"Case {i} inválido: campos 'input' e 'output' são obrigatórios.")

        parsed: dict[str, str] = {
            "input": case["input"],
            "output": case["output"],
        }

        if "label" in case:
            parsed["label"] = case["label"]

        if "grade_reduction" in case:
            parsed["grade_reduction"] = case["grade_reduction"]

        result.append(parsed)

    return result


def parse_tio2(text: str):
    pattern = re.compile(
        r"===== ?(.*?) ?=====\n"   # caso com nome
        r"(.*?)"
        r">>>>>\n"
        r"(.*?)"
        r"<<<<<",
        re.DOTALL
    )

    pattern_no_label = re.compile(
        r"=====\n"                # caso sem nome
        r"(.*?)"
        r">>>>>\n"
        r"(.*?)"
        r"<<<<<",
        re.DOTALL
    )

    cases: list[dict[str, str]] = []

    for label, inp, out in pattern.findall(text):
        cases.append({
            "label": label.strip() or None, # type: ignore
            "input": inp.strip(),
            "expected": out.strip(),
        })

    for inp, out in pattern_no_label.findall(text):
        cases.append({
            "label": None, # type: ignore
            "input": inp.strip(),
            "expected": out.strip(),
        })

    return cases


class Loader:
    # regex_tio = r"^ *#INPUT *(.*?)\n(.*?)^ *#OUTPUT *\n(.*?)^ *#END *\n?"
    regex_tio = r"^[ ]*[>]+ ?INSERT[ ]*([^\n]*?)\n([\s\S]*?)^[ ]*[=]+ ?EXPECT[ ]*\n([\s\S]*?)^[ ]*[<]+ ?FINISH[ ]*$"
    regex_tio_origin = r"^>>>>>>>> *(.*?)\n(.*?)^======== *\n(.*?)^<<<<<<<<"

    def __init__(self):
        pass

    @staticmethod
    def parse_cio(text: str, source: str):
        unit_list: list[Unit] = []
        text = "\n" + text

        pattern = r'```.*?\n(.*?)```'  # get only inside code blocks
        code = re.findall(pattern, text, re.MULTILINE | re.DOTALL)
        # join all code blocks found
        text = "\n" + "\n".join(code)

        pieces: list[dict[str, Any]] = []  # header, input, output

        open_case = False
        for line in text.splitlines():
            if line.startswith("#__case") or line.startswith("#TEST_CASE"):
                pieces.append({"header": line, "input": [], "output": []})
                open_case = True
            elif open_case:
                pieces[-1]["output"].append(line)
                if line.startswith("$end"):
                    open_case = False

        # concatenando testes contínuos e finalizando testes sem $end
        for i in range(len(pieces)):
            output = pieces[i]["output"]
            if output[-1] != "$end" and i < len(pieces) - 1:
                pieces[i + 1]["output"] = output + pieces[i + 1]["output"]
                output.append("$end")

        # removendo linhas vazias e criando input das linhas com $
        for piece in pieces:
            piece["input"] = [line[1:] for line in piece["output"] if line.startswith("$")]
            piece["output"] = [line for line in piece["output"] if line != "" and not line.startswith("#")]

        for piece in pieces:
            case = " ".join(piece["header"].split(" ")[1:])
            inp = "\n".join(piece["input"]) + "\n"
            output = "\n".join(piece["output"]) + "\n"
            unit_list.append(Unit(case, inp, output, None, source))

        return unit_list

    @staticmethod
    def parse_tio(text: str, source: str = "") -> list[Unit]:

        # identifica se tem grade e retorna case name e grade
        def parse_case_grade(value: str) -> tuple[str, None | int]:
            if value.endswith("%"):
                words = value.split(" ")
                last = value.split(" ")[-1]
                _case = " ".join(words[:-1])
                grade_str = last[:-1]           # ultima palavra sem %
                try:
                    _grade = int(grade_str)
                    return _case, _grade
                except ValueError:
                    pass
            return value, None

        matches = re.findall(Loader.regex_tio, text, re.MULTILINE | re.DOTALL)
        if len(matches) == 0:
            matches += re.findall(Loader.regex_tio_origin, text, re.MULTILINE | re.DOTALL)
        unit_list: list[Unit] = []
        for m in matches:
            case, grade = parse_case_grade(m[0])
            unit_list.append(Unit(case, m[1], m[2], grade, source))
        return unit_list

    @staticmethod
    def parse_vpl(text: str, source: str = "") -> list[Unit]:
        data_list = VplParser.parse_vpl(text)
        output: list[Unit] = []
        for m in data_list:
            output.append(Unit(m.case, m.input, m.output, m.grade, source))
        return output
    
    @staticmethod
    def parse_toml(text: str, source: str = "") -> list[Unit]:
        data_list = parse_toml(text)
        output: list[Unit] = []
        for m in data_list:
            case = m.get("label", "")
            inp = m["input"]
            outp = m["output"]
            grade = m.get("grade_reduction", None)
            output.append(Unit(case, inp, outp, grade, source))
        return output

    @staticmethod
    def parse_dir(folder: Path) -> list[Unit]:  # Updated return type to list[Unit]
        pattern_loader = PatternLoader()
        files = sorted(os.listdir(folder))
        matches = pattern_loader.get_file_sources(files)

        unit_list: list[Unit] = []  # Updated type to list[Unit]
        try:
            for m in matches:
                unit = Unit()
                unit.source = os.path.join(folder, m.label)
                unit.grade = 100
                input_file = os.path.join(folder, m.input_file)
                value = Decoder.load(input_file)
                unit.input = value + ("" if value.endswith("\n") else "\n")
                output_file = os.path.join(folder, m.output_file)
                value = Decoder.load(output_file)
                unit.set_expected(value + ("" if value.endswith("\n") else "\n"))
                unit_list.append(unit)
        except FileNotFoundError as e:
            print(str(e))
        return unit_list

    @staticmethod
    def extract_data_inside_code_fences(content: str, language: str) -> str:
        pattern = r'```' + language + r'\n(.*?)```'  # get only inside code blocks
        code = re.findall(pattern, content, re.MULTILINE | re.DOTALL)
        return "\n" + "\n".join(code)

    @staticmethod
    def parse_source(source: Path) -> list[Unit]:
        if source.is_dir():
            return Loader.parse_dir(source)
        if source.is_file():
            #  if PreScript.exists():
            #      source = PreScript.process_source(source)
            content = Decoder.load(source)
            if source.suffix == ".vpl" or source.suffix == ".cases":
                return Loader.parse_vpl(content, str(source))
            elif source.suffix == ".tio":
                return Loader.parse_tio(content, str(source))
            elif source.suffix == ".toml":
                return Loader.parse_toml(content, str(source))
            elif source.suffix == ".md":
                tests = Loader.parse_tio(content, str(source))
                tests += Loader.parse_cio(content, str(source))
                try:
                    content_fences = Loader.extract_data_inside_code_fences(content, "toml")
                    tests += Loader.parse_toml(content_fences, str(source))
                except ValueError:
                    pass
                return tests
            else:
                print("warning: target format do not supported: " + str(source))  # make this a raise
        else:
            raise FileNotFoundError('warning: unable to find: ' + str(source))
        return []



from tko.util.decoder import Decoder

class Title:
    @staticmethod
    def extract_title(readme_file: str) -> str:
        content = Decoder.load(readme_file)
        title = content.splitlines()[0]
        parts = title.split(" ")
        if parts[0].count("#") == len(parts[0]):
            del parts[0]
        title = " ".join(parts)
        return title


# from typing import override
from pathlib import Path

from tko.util.symbols import symbols
from tko.util.text import Text
from tko.game.tree_item import TreeItem
from tko.game.task_info import TaskInfo
from tko.game.task_grader import TaskGrader
import enum


class Task(TreeItem):
    str_index = "idx"

    class TaskRate(enum.Enum):
        AUTO = "auto" # rate uses % of test cases passed
        USER = "user" # rate uses user self-evaluation
        TICK = "tick" # no rate, just tick when complete

    class TaskPath(enum.Enum):
        MAIN = "main" # main task, required to complete the quest
        SIDE = "side" # side task, optional to complete the quest, usually with less xp reward

    class TaskHelp(enum.Enum):
        OPEN = "open" # help allowed
        EXAM = "exam" # no help allowed, must be done alone

    class TaskAction(enum.Enum):
        VIEW = "view" # view task details
        EDIT = "edit" # edit task details

    def __init__(self):

        super().__init__()
        self.line_number = 0
        self.line = ""
        self.info = TaskInfo()
        self.grader = TaskGrader(self.info)
        self.main_idx: int = 0
        
        self.task_rate: Task.TaskRate = Task.TaskRate.AUTO
        self.task_path: Task.TaskPath = Task.TaskPath.MAIN
        self.task_help: Task.TaskHelp = Task.TaskHelp.OPEN
        self.task_action: Task.TaskAction = Task.TaskAction.EDIT

        self.skills: dict[str, int] = {} # skills
        
        self.xp: int = 1
        
        self.target = ""
        self.quest_key = ""
        self.remote_name = ""
        self.__origin_folder: Path | None = None
        self.__workspace_folder: Path | None = None
        self.__is_reachable = False
        self.default_min_value = 5 # default min grade to complete task

    def get_full_title(self, key_pad: None | int):
        if key_pad is None:
            key_pad = len(self.get_key())
        if not f"@{self.get_key()}" in self.get_title():
            return f"@{self.get_key().ljust(key_pad)} {self.get_title()}"
        return self.get_title()

    def set_reachable(self, reachable: bool):
        self.__is_reachable = reachable
        return self
    
    def is_optional(self):
        return self.task_path == Task.TaskPath.SIDE
    
    def is_leet(self):
        return self.task_rate == Task.TaskRate.AUTO

    def is_reachable(self) -> bool:
        return self.__is_reachable

    def is_link(self):
        if self.task_action == Task.TaskAction.VIEW:
            return True
        return self.__origin_folder is None and self.__workspace_folder is None
    
    def set_remote_view_type(self):
        self.__origin_folder = None
        self.__workspace_folder = None
        return self

    def is_import_type(self):
        return self.task_action == Task.TaskAction.EDIT and self.__origin_folder is not None and self.__workspace_folder is not None
    
    def is_static_type(self):
        if self.is_link():
            return False
        return self.get_origin_folder() == self.get_workspace_folder()

    def set_origin_folder(self, folder: Path):
        self.__origin_folder = folder
        return self
    
    def get_origin_readme(self) -> Path:
        origin_folder = self.get_origin_folder()
        if origin_folder is not None:
            return origin_folder / "README.md"
        return Path("")
    
    def set_workspace_folder(self, folder: Path):
        self.__workspace_folder = folder
        return self

    def get_origin_folder(self) -> Path | None:
        return self.__origin_folder
    
    def get_workspace_folder(self) -> Path | None:
        if self.__workspace_folder is not None:
            return self.__workspace_folder
        return self.__origin_folder
    
    def decode_from_dict(self, value: str):
        value_list = value[1:-1]
        kv_dict: dict[str, str] = {}
        key_values = value_list.split(",")
        for kv in key_values:
            if ":" not in kv:
                continue
            (k, val) = kv.split(":")
            kv_dict[k.strip()] = val.strip()
        
        self.info.load_from_kv(kv_dict)
        if self.str_index in kv_dict:
            self.main_idx = int(kv_dict[self.str_index])

        # deprecated
        # for k, val in kv_dict.items():
        #     if k == "cov" or k == 'cove':
        #         self.info.rate = int(val)
        #     elif k == "aut" or k == "appr":
        #         self.info.flow = int(val)
        #     elif k == "hab" or k == "auto":
        #         self.info.edge = int(val)
        #     elif k == "desc":
        #         self.info.neat = int(val)
        #     elif k == "desire":
        #         self.info.cool = int(val)
        #     elif k == "effort":
        #         self.info.easy = int(val)

    def load_from_db(self, value: str):
        if value.startswith("{"):
            self.decode_from_dict(value)
        else:
            raise ValueError(f"Invalid task value format: {value}. Expected format is 'flow:edge:main_idx:rate' or '{self.str_index}:value'.")

    def save_to_db(self) -> str:
        kv_dict = self.info.get_kv()
        if self.main_idx != 0:
            kv_dict[self.str_index] = str(self.main_idx)
        return "{" + ", ".join(f"{k}:{v}" for k, v in kv_dict.items()) + "}"

    def is_db_empty(self) -> bool:
        return len(self.info.get_kv()) == 0

    def get_prog_color(self, value: int, min_value: None | int = None) -> str:
        if min_value is None:
            min_value = self.default_min_value
        prog = value
        if prog == 0:
            return "c"
        if prog < min_value:
            return "r"
        if prog < 10:
            return "y"
        if prog == 10:
            return "g"
        return "w"  

    def get_prog_symbol(self, value: int, min_value: None | int = None) -> Text:
        if min_value is None:
            min_value = self.default_min_value
        color = self.get_prog_color(value, min_value)
        prog = value
        if prog == 0:
            return Text().add("x")
        if prog < min_value:
            return Text().addf(color, str(prog))
        if prog < 10:
            return Text().addf(color, str(prog))
        if prog == 10:
            return Text().addf(color, symbols.check.text)
        return Text().add("0")

    def get_xp(self) -> int:
        if self.xp == 0:
            return 1
        return self.xp

    def get_percent(self) -> float:
        value = self.grader.get_percent()
        if value < 0.1:
            return 0.0
        return value

    def get_ratio(self) -> float:
        return self.grader.get_ratio()

    def is_complete(self):
        return self.grader.get_percent() >= 70

    def not_started(self):
        return self.grader.get_percent() == 0

    def in_progress(self):
        return 0 < self.grader.get_percent() < 100

    # @override
    def __str__(self):
        lnum = str(self.line_number).rjust(3)
        key = "" if self.get_full_key() == self.get_title() else self.get_full_key() + " "
        return f"{lnum} key:{key} title:{self.get_title()} skills:{self.skills} remote:{self.target}"

    def has_at_symbol(self):
        return any([s.startswith("@") for s in self.get_title().split(" ")])



from pathlib import Path
import subprocess
import os
from subprocess import PIPE


class Runner:
    def __init__(self):
        pass

    @staticmethod
    def subprocess_run(cmd: str | list[str], input_data: str = "", timeout: None | float = None, folder: Path | None = None, shell_mode: bool = False) -> tuple[int, str, str]:
        try:
            env = os.environ.copy()
            env['NO_COLOR'] = '1'
            env['FORCE_COLOR'] = '0'

            answer = subprocess.run(cmd, 
                                    cwd=folder, 
                                    env=env, 
                                    shell= shell_mode, 
                                    input=input_data, 
                                    stdout=PIPE, 
                                    stderr=PIPE, 
                                    timeout=timeout, 
                                    text=True
                                    )
            err = ""
            if answer.returncode != 0:
                err = answer.stderr + Runner.decode_code(answer.returncode)
            return answer.returncode, answer.stdout, err
        except subprocess.TimeoutExpired:
            err = "fail: processo abortado depois de {} segundos".format(timeout)
            return 1, "", err

    @staticmethod
    def clear_screen():
        if os.name == 'nt':
            os.system('cls')
        else:
            os.system('clear')

    @staticmethod
    def decode_code(return_code: int) -> str:
        code = 128 - return_code
        if code == 127:
            return ""
        if code == 139:
            return "fail: segmentation fault"
        if code == 134:
            return "fail: runtime exception"
        return "fail: execution error code " + str(code)



from tko.game.quest import Quest
from tko.game.task import Task
# from tko.game.graph import Graph

from icecream import ic # type: ignore
from tko.play.floating_grade import FloatingGrade
from tko.play.floating_input_text import FloatingInputText
from tko.settings.settings import Settings
from tko.play.tracker import Tracker
from tko.util.text import Text
from tko.cmds.cmd_down import CmdDown
from tko.cmds.cmd_run import Run
from tko.settings.rep_source import RepSource

from tko.util.param import Param

from tko.cmds.cmd_down import DownActions

from tko.play.fmt import Fmt
from tko.play.floating import Floating
from tko.play.gui import Gui
from tko.play.opener import Opener

from tko.play.tasktree import TaskAction
from typing import Callable

from tko.down.drafts import Drafts
import os

from tko.logger.log_item_self import LogItemSelf
from tko.logger.log_item_move import LogItemMove

import shutil
import tempfile
import subprocess
from pathlib import Path


class PlayActions:

    def __init__(self, gui: Gui):
        self.app = Settings().app
        self.settings = Settings()
        self.fman = gui.fman
        self.rep = gui.rep
        self.tree = gui.tree
        self.game = gui.game
        self.graph_opened: bool = False
        self.gui = gui

        
    def open_link_without_stdout_stderr(self, link: str):
        if link.startswith("http://") or link.startswith("https://"):
            outfile = tempfile.NamedTemporaryFile(delete=False)
            subprocess.Popen("python3 -m webbrowser -t {}".format(link), stdout=outfile, stderr=outfile, shell=True)
        else:
            Opener(Settings()).open_files([link])

    @staticmethod
    def get_task_folder(task: Task) -> Path:
        folder = task.get_workspace_folder()
        if folder is None:
            return Path("")
        return folder

    def reload_game(self):
        self.rep.load_game(try_update=False, silent=True)

    def delete_folder_ask(self):
        def delete_folder(text: str):
            obj = self.tree.get_selected_throw()
            if not isinstance(obj, Task):
                return
            if obj.get_key() != text:
                self.fman.add_input(
                    Floating().bottom().right()
                    .put_text("\nTexto digitado não corresponde ao identificador da tarefa.\n")
                    .set_error()
                )
                return
            folder = self.get_task_folder(obj)
            try:
                shutil.rmtree(folder)
                self.fman.add_input(
                    Floating().bottom().right()
                    .put_text(f"\nPasta {folder} apagada com sucesso.\n")
                    .set_warning()
                )
            except OSError as e:
                self.fman.add_input(
                    Floating().bottom().right()
                    .put_text(f"\nErro ao apagar a pasta {folder}: {e}\n")
                    .set_error()
                )
            if obj.get_remote_name() == RepSource.STUDENT_SANDBOX_NAME:
                self.tree.move_down()
            self.rep.load_game(try_update=False, silent=True)

        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            folder = self.get_task_folder(obj)
            if folder == Path("") or not os.path.exists(folder):
                self.fman.add_input(
                    Floating().bottom().right()
                    .put_text("\nEssa tarefa não possui pasta de código local.\n")
                    .set_error()
                )
                return
            self.fman.add_input(FloatingInputText(Text().add(f"Para apagar essa pasta, digite ").addf("y", f"{obj.get_key()}"), action=delete_folder))
        else:
            self.fman.add_input(
                Floating().bottom().right()
                .put_text("\nVocê só pode apagar pastas de tarefas.\n")
                .set_error()
            )

    def open_code(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            folder = self.rep.get_task_folder_for_label(task.get_full_key())
            if os.path.exists(folder):
                opener = Opener(self.settings).set_fman(self.fman)
                opener.set_target([folder]).set_language(self.rep.data.get_lang())
                opener.load_folders_and_open()
            else:
                self.fman.add_input(
                     Floating().bottom().right()
                    .put_text("\nO arquivo de código não foi encontrado.\n")
                    .set_error()
                )
        else:
            self.fman.add_input(
                 Floating().bottom().right()
                .put_text("\nVocê só pode abrir o código")
                .put_text("de tarefas baixadas.\n")
                .set_error()
            )

    def open_link(self):
        obj = self.tree.get_selected_throw()
        if isinstance(obj, Task):
            task: Task = obj
            if task.is_link():
                try:
                    self.open_link_without_stdout_stderr(task.target)
                except Exception as _:
                    pass
            self.fman.add_input(
                 Floating().bottom().right()
                .set_header(" Abrindo link ")
                .put_text("\n " + task.target + " \n")
                .set_warning()
            )
        elif isinstance(obj, Quest):
            self.fman.add_input(
                 Floating().bottom().right()
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .set_error()
            )
        else:
            self.fman.add_input(
                 Floating().bottom().right()
                .put_text("\nEsse é um grupo.")
                .put_text("\nVocê só pode abrir o link")
                .put_text("de tarefas.\n")
                .set_error()
            )

    def register_action(self, task: Task):
        self.rep.logger.store(LogItemSelf().set_task(task))

    def self_evaluate(self):
        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Task):
            task: Task = obj
            if ic.enabled:
                if not task.info.feedback:
                    task.info.rate = 100
                    task.info.feedback = True
                else:
                    task.info.rate = 0
                    task.info.feedback = False
            else:
                if task.is_link() or task.task_rate == Task.TaskRate.TICK:
                    if task.info.feedback:
                        task.info.feedback = False
                        task.info.rate = 0
                        self.rep.logger.store(LogItemSelf().set_task(task))
                    else:
                        task.info.feedback = True
                        task.info.rate = 100
                        self.rep.logger.store(LogItemSelf().set_task(task))
                else:
                    self.fman.add_input(
                        FloatingGrade(obj, lambda task: self.rep.logger.store(LogItemSelf().set_task(task)))
                    )
            return

    def create_draft(self):
        local_source = self.game.get_sandbox_source()
        workspace_folder: Path = local_source.get_source_workspace() / local_source.get_default_sandbox_draft_folder()
        if not os.path.exists(workspace_folder):
            os.makedirs(workspace_folder)
        current_folders_on_rep = os.listdir(workspace_folder)
        
        def __create(draft_name: str):
            folder:Path = workspace_folder / draft_name
            if not folder.exists():
                folder.mkdir()
            else:
                self.fman.add_input(
                    Floating().bottom().right()
                    .put_text(f"\nA pasta {folder} já existe.\n")
                    .set_error()
                )
                return

            draft = ""
            if self.rep.data.lang in Drafts.drafts:
                draft = Drafts.drafts[self.rep.data.lang]
            with open(folder / f"draft.{self.rep.data.lang}", "w", encoding="utf-8") as f:
                f.write(draft)

            task_keys_only: list[str] = []
            for quest in self.game.quests.values():
                for task in quest.get_tasks():
                    task_keys_only.append(task.get_key())
            draft_id = Drafts.find_max_numbered_key(task_keys_only=task_keys_only) + 1
            key = Drafts.create_draft_key(draft_id)
            Drafts.create_sandbox_draft(folder, key)
            
            self.tree.selected_item = f"{RepSource.STUDENT_SANDBOX_NAME}@{key}"
            self.tree.expanded.add(f"{RepSource.STUDENT_SANDBOX_NAME}@{RepSource.DEFAULT_DRAFT_FOLDER}")
            self.rep.data.selected = self.tree.selected_item
            self.rep.load_game(try_update=False, silent=True) # recarrega o jogo
            self.fman.add_input( Floating().bottom().right()
                                .put_text(f"Rascunho criado em {folder}")
                                .put_text("\nVocê pode renomear a pasta manualmente e apertar shift R para recarregar.")
                                .set_warning())

        self.fman.add_input(FloatingInputText(Text().add("Nome do rascunho"), __create, current_folders_on_rep))

    def down_remote_task(self):

        obj = self.tree.get_selected_throw()
        
        if isinstance(obj, Quest):
            self.fman.add_input(
                 Floating().bottom().right()
                .put_text("\nEssa é uma missão.")
                .put_text("\nVocê só pode baixar tarefas.\n")
                .set_error()
            )
            return
        if not isinstance(obj, Task):
            return
        self.__down_remote_task(obj)
    
    def __down_remote_task(self, task: Task) -> None:
        if not task.is_import_type():
            self.fman.add_input(
                 Floating().bottom().right().put_text("\nEssa não é uma tarefa de baixável.\n").set_error()
            )
            return
        down_frame = (
             Floating().bottom().right().set_warning().set_text_ljust().set_header(" Baixando tarefa ")
        )
        # down_frame.put_text(f"\ntko down {self.rep.alias} {task.key} -l {lang}\n")
        self.fman.add_input(down_frame)

        def fnprint(text: str):
            down_frame.put_text(" " + text)
            down_frame.draw()
            Fmt.refresh()

        cmd_down = CmdDown(repo=self.rep, task_key=task.get_full_key(), settings=self.settings)
        cmd_down.set_fnprint(fnprint)
        result = cmd_down.execute()
        if result:
            self.rep.logger.store(
                LogItemMove().set_key(task.get_full_key()).set_mode(LogItemMove.Mode.DOWN)
            )


    def select_task_action(self, task: Task) -> None:
        _, action = self.tree.get_task_action(task)
        if action == TaskAction.BAIXAR:
            self.__down_remote_task(task)
            return
        if action == TaskAction.VISITAR:
            self.open_link()
            return
        folder = task.get_workspace_folder()
        if not folder:
            raise Warning("Folder não encontrado")
        self.run_selected_task(task, folder)


    def open_versions(self):
        try:
            obj = self.tree.get_selected_throw()

            if isinstance(obj, Task):
                task: Task = obj
                track_folder = self.rep.paths.get_track_task_folder(task.get_full_key())
                tracker = Tracker()
                tracker.set_folder(track_folder)
                if task.get_full_key() in self.rep.logger.tasks.task_dict:
                    log_sort = self.rep.logger.tasks.task_dict[task.get_full_key()]

                    msg, folder = tracker.unfold_files(log_sort)
                    cmd = self.settings.app.get_editor()
                    fullcmd = "{} {}".format(cmd, folder)
                    outfile = tempfile.NamedTemporaryFile(delete=False)
                    subprocess.Popen(fullcmd, stdout=outfile, stderr=outfile, shell=True)
                    self.fman.add_input(
                        Floating().bottom().right()
                        .put_text("\nAs versões da tarefa foram descompactadas em uma pasta temporária")
                        .put_text(msg)
                        .put_text(f"Abrindo com o comando: {fullcmd}\n")
                    )
        except:
            pass

    def select_task(self) -> Callable[[], None]:
        try:
            obj = self.tree.get_selected_throw()

            if isinstance(obj, Quest):
                self.tree.toggle(obj)
                return lambda: None

            if isinstance(obj, Task):
                task: Task = obj
                return lambda: self.select_task_action(task)
        except IndexError as _:
            pass

        return lambda: None
        
    def run_selected_task(self, task: Task, task_dir: Path) -> None:
        folder = task_dir
        run = Run(settings=self.settings, target_list=[folder], param=Param.Basic())
        run.set_lang(self.rep.data.lang)
        opener = Opener(self.settings).set_language(self.rep.data.get_lang()).set_target([folder])
        run.set_opener(opener)
        run.set_run_without_ask(False)
        run.set_curses(True)
        run.set_task(self.rep, task)
        run.build_wdir()
        
        if not run.wdir.has_solver():
            lang = self.rep.data.get_lang()
            draft_folder = folder / lang
            draft_path = DownActions(self.settings).create_default_draft(draft_folder, self.rep.data.get_lang())
            msg =  Floating().bottom().right().set_warning()
            msg.put_text("\nNenhum arquivo de código na linguagem {} encontrado.".format(self.rep.data.get_lang()))
            msg.put_text("\nUm arquivo de rascunho vazio foi criado em\n {} ".format(draft_path))
            self.fman.add_input(msg)
        run.execute()


from __future__ import annotations
import os
from typing import Any
from tko.settings.app_settings import AppSettings
from platformdirs import user_data_dir
from pathlib import Path
import yaml #type: ignore

from tko.settings.user_languages import UserLanguages
from tko.util.text import Text
from tko.play.colors import Colors
from tko.util.decoder import Decoder
from tko.settings.repository import remove_git_merge_tags

def singleton(class_): # type: ignore
    instances = {}
    def getinstance(*args, **kwargs): # type: ignore
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_] # type: ignore
    return getinstance # type: ignore

@singleton
class Settings:
    CFG_FILE = "settings.yaml"
    class Defaults:
        alias_git = {
            "poo": "https://github.com/qxcodepoo/arcade.git",
            "fup": "https://github.com/qxcodefup/arcade.git",
            "ed": "https://github.com/qxcodeed/arcade.git"
        }

    def __init__(self):
        self.__gitrepos = "gitrepos"
        self.__appcfg = "appcfg"
        self.__colors = "colors"

        self.package_name = 'tko'
        self.dict_alias_git: dict[str, str] = {}
        self.app = AppSettings()
        self.languages: UserLanguages = UserLanguages(Path("")).set_default_path().load_file_settings()
        self.colors = Colors()

        self.settings_file: str = ""
        self.data: dict[str, Any] = {}

    def set_settings_file(self, path: str):
        self.settings_file = path
        return self

    def get_settings_file(self) -> str:
        if self.settings_file == "":
            self.settings_file = os.path.join(user_data_dir(self.package_name), self.CFG_FILE) # type: ignore
        if not os.path.exists(self.settings_file):
            os.makedirs(os.path.dirname(self.settings_file), exist_ok=True)
        return self.settings_file

    def reset(self):
        self.dict_alias_git = self.Defaults.alias_git.copy()
        self.app = AppSettings()
        self.colors = Colors()
        self.languages.reset().save_file_settings()

        return self


    def set_alias_git(self, alias: str, git_url: str):
        self.dict_alias_git[alias] = git_url
        return self

    def get_alias_git(self, alias: str) -> str:
        if alias in self.dict_alias_git:
            return self.dict_alias_git[alias]
        raise Warning(f"Repositório git label {alias} não encontrado")

    def has_alias_git(self, alias: str) -> bool:
        return alias in self.dict_alias_git

    def load_settings(self):
        try:
            settings_file = self.get_settings_file()


            content = Decoder.load(settings_file)

            lines = content.splitlines()
            lines = remove_git_merge_tags(lines)
            data: Any = yaml.safe_load("\n".join(lines))

            if data is None or not isinstance(data, dict):
                raise FileNotFoundError(f"Arquivo de configuração vazio: {settings_file}")
            self.data = data
            self.dict_alias_git = data.get(self.__gitrepos, self.Defaults.alias_git) # type: ignore
            if len(self.dict_alias_git.keys()) == 0: # type: ignore
                self.dict_alias_git = self.Defaults.alias_git.copy()
            self.app = AppSettings().from_dict(data.get(self.__appcfg, AppSettings())) # type: ignore
            self.colors = Colors().from_dict(data.get(self.__colors, Colors())) # type: ignore
            self.languages.load_file_settings()
        except:
            self.reset()
            self.save_settings()
        return self
    
    def save_settings(self):
        file = self.get_settings_file()
        value: dict[str, Any] = {
            self.__gitrepos: self.dict_alias_git,
            self.__appcfg: self.app.to_dict(),
            self.__colors: self.colors.to_dict()
        }
        with open(file, "w", encoding="utf-8") as f:
            yaml.dump(value, f)
        self.languages.save_file_settings()
        return self

    def __str__(self):
        output: list[str] = []
        output.append(str(Text.format("{g}", "Arquivo global configuração:")))
        output.append("- " + self.get_settings_file())
        output.append("- " + self.languages.path.as_posix())
        output.append("")
        output.append(str(Text.format("{g}", "Fontes de tarefas remotas cadastradas:")))
        max_alias = max([len(key) for key in self.dict_alias_git])
        for key in self.dict_alias_git:
            output.append("- @{} : {}".format(key.ljust(max_alias), self.dict_alias_git[key]))
        output.append("")
        
        app_str = str(self.app)
        return "\n".join(output) + "\n\n" + app_str



from tko.util.text import Text
from tko.util.symbols import symbols

class Visual:
    def get_lr(self, test: bool) -> tuple[str, str]:
        if test:
            return symbols.right_arrow_filled.text, symbols.left_arrow_filled.text
        else:
            return " ", " "

    def render_button(self, info: str, test: bool):
        left, right = self.get_lr(test)
        bg = "X" if test else ""
        return Text().addf(bg, left).addf(bg, info).addf(bg, right)


from __future__ import annotations
from tko.logger.log_item_base import LogItemBase

import enum

class LogItemMove(LogItemBase):
    class Mode(enum.Enum):
        NONE = "NONE"
        DOWN = "DOWN"
        PICK = "PICK"
        BACK = "BACK"
        EDIT = "EDIT"

    mode_str = "mode"
    def __init__(self):
        super().__init__(LogItemBase.Type.MOVE)
        self.mode: LogItemMove.Mode = LogItemMove.Mode.NONE

    def set_mode(self, action: LogItemMove.Mode):
        self.mode = action
        return self

    def encode_line(self) -> str:
        return f'{super().encode_line()}, {self.mode_str}:{self.mode.value}'

    def identify_kv(self, kv: dict[str, str]) -> bool:
        if self.mode_str in kv:
            self.mode = LogItemMove.Mode(kv[self.mode_str])
        if self.mode == LogItemMove.Mode.NONE:
            return False
        return True



from tko.game.task import Task
from tko.play.floating import FloatingABC, Floating
from tko.util.text import Text
from tko.game.task_grader import TaskGrader
from tko.util.symbols import symbols

# import ABC, abstractmethod
from abc import ABC, abstractmethod
from typing import Callable

# from typing import override

import curses


class InputLine(ABC):

    @abstractmethod
    def send_key(self, key: int) -> None:
        pass

    @abstractmethod
    def set_focus(self, focus: bool) -> None:
        pass

    @abstractmethod
    def get_text(self) -> Text:
        pass

class InputSlide(InputLine):
    def __init__(self, prefix: Text, opt_msgs: list[tuple[str, Text]], index: int = 0):
        self.prefix = prefix
        self.opt_msgs = opt_msgs
        self.index: int = index
        self.focus: bool = False

    def send_key(self, key: int) -> None:
        size = len(self.opt_msgs)
        if key == curses.KEY_LEFT:
            self.index = max(0, self.index - 1)
        elif key == curses.KEY_RIGHT:
            self.index = min(size - 1, self.index + 1)
        elif key == ord("-"):
            self.index = 0
        elif key == ord("+") or key == ord("="):
            self.index = size - 1

    def set_focus(self, focus: bool) -> None:
        self.focus = focus

    def get_text(self) -> Text:
        color = "Y" if self.focus else ""
        text = Text().addf(color, self.prefix).add(" ")
        for i, c in enumerate(self.opt_msgs):
            opt, _ = c
            text.addf("Y" if i == self.index else "", opt)
        text.add(" ├").add(self.opt_msgs[self.index][1])
        return text
    
class InputText(InputLine):
    def __init__(self, prompt: Text, text: str = ""):
        self.prompt = prompt
        self.text = text
        self.focus = False
        self.number_only = False

    def send_key(self, key: int) -> None:
        if key == curses.KEY_BACKSPACE:
            if len(self.text) > 0:
                self.text = self.text[:-1]
        elif 32 <= key <= 126:
            if self.number_only:
                if chr(key).isdigit():
                    self.text += chr(key)
            else:
                info = chr(key)
                if info == "," or info == " " or info == ".":
                    self.text += info
                elif chr(key).isalpha() or chr(key).isdigit() or chr(key) == "_":
                    self.text += info

    def set_focus(self, focus: bool) -> None:
        self.focus = focus

    def set_number_only(self, number_only: bool):
        self.number_only = number_only
        return self

    def get_text(self) -> Text:
        data = Text().addf("Y" if self.focus else "", self.prompt).add(" ")
        data = data.add("├ ").addf("g",self.text).add(symbols.cursor if self.focus else "")
        return data
    
class InputBoolean(InputLine):
    def __init__(self, prefix: Text, start_value: str):
        self.prefix = prefix
        self.value = start_value
        self.focus = False

    def send_key(self, key: int) -> None:
        if key == curses.KEY_LEFT:
            self.value = "0"
            self.focus = False
        elif key == curses.KEY_RIGHT:
            self.value = "1"
            self.focus = False

    def set_focus(self, focus: bool) -> None:
        self.focus = focus
        if focus and self.value == "":
            self.value = "0"

    def get_text(self) -> Text:
        color = "Y" if self.focus else ""
        text = Text().addf(color, self.prefix).add(" │ ")
        text.addf("Y" if self.value == "0" else "", "Não").add(" ").addf("Y" if self.value == "1" else "", "Sim")
        return text


class FloatingGrade(FloatingABC):
    def __init__(self, task: Task, fn_exit: Callable[[Task], None] | None = None):
        self.floating = Floating()
        self._task = task
        self._grader = TaskGrader(task.info)
        self._line = 0 if not task.is_leet() else 1
        self.floating.set_text_ljust()
        self.floating.frame.set_border_color("g")
        self.floating.set_header_text(Text.format("{y/}", " Utilize os direcionais e texto para marcar"))
        self.floating.set_footer_text(Text.format("{y/}", " Pressione Enter para confirmar, Esc para cancelar"))
        self.fn_exit = fn_exit

        progression: list[tuple[str, Text]] = [
            ("x", Text().addf("g", " Nada")),
            ("1", Text().addf("y", " 10%")),
            ("2", Text().addf("y", " 20%")),
            ("3", Text().addf("y", " 30%")),
            ("4", Text().addf("y", " 40%")),
            ("5", Text().addf("y", " 50%")),
            ("6", Text().addf("y", " 60%")),
            ("7", Text().addf("y", " 70%")),
            ("8", Text().addf("y", " 80%")),
            ("9", Text().addf("y", " 90%")),
            ("✓", Text().addf("y", " 100%"))]


        if self._task.is_leet():
            texto_leet = "Testes que passaram na última execução: "
        else:
            texto_leet = "Qual percentual da atividade você fez?  "
        
        guided = "" if not self._task.info.feedback else ("1" if self._task.info.guided else "0")
        concept = "" if not self._task.info.feedback else ("1" if self._task.info.ia_concept else "0")
        problem = "" if not self._task.info.feedback else ("1" if self._task.info.ia_problem else "0")
        code = "" if not self._task.info.feedback else ("1" if self._task.info.ia_code else "0")
        debug = "" if not self._task.info.feedback else ("1" if self._task.info.ia_debug else "0")
        refactor = "" if not self._task.info.feedback else ("1" if self._task.info.ia_refactor else "0")
        
        self.rate_slide     = InputSlide(Text().add(texto_leet), progression, self._task.info.rate // 10)
        self.human_text     =     InputText(Text().add("Usou ajuda de monitor, amigo? Se sim, qual nome?    "), self._task.info.friend)
        self.time_text      =     InputText(Text().add("Tempo total estimado, estudo + código, em minutos?  "), str(self._task.info.study)).set_number_only(True)
        self.guide_text     =  InputBoolean(Text().add("Fez seguindo a aula presencial ou vídeo aula?       "), guided)
        self.iaconcept_text =  InputBoolean(Text().add("ESTUDAR conceitos sem gerar a solução do problema?  "), concept)
        self.iaproblem_text =  InputBoolean(Text().add("ENTENDER o problema a ser resolvido?                "), problem)
        self.iacode_text    =  InputBoolean(Text().add("GERAR ou CORRIGIR código relacionado ao problema?   "), code)
        self.iadebug_text   =  InputBoolean(Text().add("COMPREENDER mensagens de ERRO ou SAÍDA incorreta?   "), debug)
        self.iarefactor_text=  InputBoolean(Text().add("REFATORAR o código só após fazer tudo sozinho?      "), refactor)
        self.input_lines: list[InputLine] = [
            self.rate_slide, 
            self.human_text, 
            self.time_text,
            self.guide_text, 
            self.iaconcept_text, 
            self.iaproblem_text, 
            self.iacode_text, 
            self.iadebug_text,
            self.iarefactor_text
        ]

    def set_focus(self):
        for i, line in enumerate(self.input_lines):
            line.set_focus(i == self._line)

    def set_exit_fn(self, fn: Callable[[], None]):
        self.floating.exit_fn = fn
        return self

    def is_enable(self) -> bool:
        return self.floating.enable

    def update_content(self):
        self.set_focus()

        content = self.floating.content
        content.clear()
        content.append(Text().add("         Pontue de acordo com a última fez que você (re)fez a tarefa do zero (sprint)         "))
        width = 90
        somatorio = (Text() .addf('g', f'{round(self._task.get_percent()):>3}%'))

        content.append(Text().add("╔") + Text().add(" Tarefa:").add(somatorio).center(width, Text.Token("═", "")))
        enabled_option = "╠═ "
        op_prefix = enabled_option if not self._task.is_leet() else "║  "
        content.append(Text().add(op_prefix).add(self.rate_slide.get_text()))
        content.append(Text().add(enabled_option).add(self.human_text.get_text()))
        content.append(Text().add(enabled_option).add(self.time_text.get_text()))
        content.append(Text().add(enabled_option).add(self.guide_text.get_text()))
        content.append(Text().add("╠═") + Text().add(" Você usou IA (LLMs) para ").center(width, Text.Token("═", "")))
        content.append(Text().add(enabled_option).add(self.iaconcept_text.get_text()))
        content.append(Text().add(enabled_option).add(self.iaproblem_text.get_text()))
        content.append(Text().add(enabled_option).add(self.iacode_text.get_text()))
        content.append(Text().add(enabled_option).add(self.iadebug_text.get_text()))
        content.append(Text().add(enabled_option).add(self.iarefactor_text.get_text()))
        content.append(Text().add("╚═══════════════════════════════════════════════════════════").ljust(width, Text.Token("═", "")))


    def draw(self):
        self.update_content()
        self.floating.draw()

    def change_task(self):
        if not self._task.is_leet():
            self._task.info.rate = self.rate_slide.index * 10
        self._task.info.feedback = True
        self._task.info.friend = self.human_text.text
        self._task.info.guided = self.guide_text.value == "1"
        self._task.info.ia_concept = self.iaconcept_text.value == "1"
        self._task.info.ia_code = self.iacode_text.value == "1"
        self._task.info.ia_debug = self.iadebug_text.value == "1"
        self._task.info.ia_problem = self.iaproblem_text.value == "1"
        self._task.info.ia_refactor = self.iarefactor_text.value == "1"
        self._task.info.set_study(self.time_text.text)

    def send_key_up(self):
        if not(self._task.is_leet() and self._line == 1):
            self._line = max(self._line - 1, 0)

    def send_key_down(self):
        self._line = min(self._line + 1, len(self.input_lines) - 1)

    def send_key(self, key: int):
        self.input_lines[self._line].send_key(key)
        # self.change_task()
    
    def process_input(self, key: int) -> int:

        if key == curses.KEY_UP:
            self.send_key_up()
        elif key == curses.KEY_DOWN or key == ord("\t"):
            self.send_key_down()
        elif key == ord('\n'):
            if self._line != len(self.input_lines) - 1:
                self.send_key_down()
            else:
                self.floating.enable = False
                self.change_task()
                if self.fn_exit is not None:
                    self.fn_exit(self._task)
        elif key == curses.KEY_EXIT:
            self.floating.enable = False
        else:
            self.send_key(key)
        return -1


import os
from pathlib import Path
from tko.play.language_setter import LanguageSetter
from tko.util.text import Text
from tko.settings.repository import Repository
from tko.settings.rep_paths import RepPaths
import shutil
class RepStarter:
    def __init__(self, folder: Path | None, language: str | None = None):
        # if folder is set, use folder, else use local folder.
        self.folder: Path = Path.cwd()
        if folder is not None:
            self.folder = folder
        self.language = language

    def execute(self) -> bool:
        repo = self.create_repository()
        if repo is None:
            return False
        
        self.repo = repo
        self.create_empty_rep()
        # erase cache folder to avoid conflicts
        cache_folder = repo.paths.get_cache_folder()
        if os.path.exists(cache_folder):
            shutil.rmtree(cache_folder)
        os.makedirs(cache_folder, exist_ok=True)
        if self.language is not None:
            repo.data.lang = self.language
            print(Text.format("A linguagem do repositório foi definida como {y}.", self.language))
        else:
            LanguageSetter.check_lang_in_text_mode(self.repo)
        
        repo.save_config()
        return True

    def print_end_msg(self):
        print(Text.format("Voce pode acessar o repositório com o comando {g} {y}", "tko open"))
    
    def create_repository(self) -> Repository | None:
        path_parents = RepPaths.rec_search_for_repo_parents(self.folder)

        if path_parents is not None and path_parents.resolve() == self.folder.resolve():
            print(Text.format("Já existe um repositório TKO na pasta {y}", self.folder.resolve()))
            print(Text.format("Deseja resetar o repositório? ({g}/{r}): ", "s", "n"), end="")
            op = input()
            if op == "n":
                return None

        elif path_parents is not None:
            if self.folder != path_parents:
                print(Text.format("Você está tentando criar um repositório dentro de outro, pois já existe rep em {r}.", path_parents))
                print(Text.format("Você pode apagar o repositório antigo, criar seu repositório em outro lugar ou sobrescrever as configurações."))
            self.folder = path_parents
            print(Text.format("Deseja sobrescrever as configurações do repositório em {y} ? ({g}/{r}): ", self.folder, "s", "n"), end="")
            op = input()
            if op == "n":
                return None
        else:
            path_subdir_list = RepPaths.rec_search_for_repo_subdir(self.folder)
            if len(path_subdir_list) > 0:
                print(Text.format("Você está tentando criar um repositório TKO na pasta {y}", self.folder.resolve()))
                print(Text.format("Porém já existem repositórios TKO abaixo dessa pasta. Mova ou apague-os"))
                for path in path_subdir_list:
                    print(Text.format("- {r}", path))
                return None

        return Repository(self.folder)
    
    def create_empty_rep(self):
        source = self.repo.get_student_sandbox()
        self.repo.data.set_source(source)
        folder = source.get_source_workspace()
        print(f"Criando repositório vazio, utilizando a pasta {folder} como pasta para atividades locais")
    


from __future__ import annotations
from tko.settings.task_listener import TaskListener
from tko.settings.daily_listener import DailyListener
from tko.logger.log_history import LogHistory
from tko.logger.log_item_base import LogItemBase
from pathlib import Path


class Logger:

    def __init__(self, rep_folder: Path):
        self.tasks = TaskListener()
        self.daily = DailyListener()
        self.history = LogHistory(rep_folder, [self.tasks.handle_log_entry, self.daily.handle_entry_incoming])
    
    def store(self, action: LogItemBase):
        self.history.append_new_action(action)
